# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from dataclasses import dataclass, field
from typing import Dict

import numpy as np
from pydantic import ConfigDict

from qat.backend.passes.purr.analysis import TriageResult
from qat.backend.qblox.config.specification import SequencerConfig
from qat.backend.qblox.ir import Sequence
from qat.executables import AcquireData, ChannelExecutable
from qat.model.device import PulseChannel


@dataclass
class QbloxPackage:
    sequence: Sequence = None
    sequencer_config: SequencerConfig = field(default_factory=lambda: SequencerConfig())
    timeline: np.ndarray = None


class QbloxExecutable(ChannelExecutable):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    packages: Dict[PulseChannel, QbloxPackage]
    triage_result: TriageResult

    @property
    def acquires(self) -> list[AcquireData]:
        return []



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from qat.pipelines.updateable import PipelineConfig as PipelineConfig

from .purr.waveform_v1 import EchoPipeline as EchoPipeline
from .purr.waveform_v1 import echo8 as echo8
from .purr.waveform_v1 import echo16 as echo16
from .purr.waveform_v1 import echo32 as echo32



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import abc
from typing import Optional

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.ir.conversion import ConvertToPydanticIR
from qat.middleend.passes.analysis import PydActivePulseChannelAnalysis
from qat.middleend.passes.purr.analysis import (
    ActivePulseChannelAnalysis,
)
from qat.middleend.passes.purr.transform import (
    AcquireSanitisation,
    BatchedShots,
    EndOfTaskResetSanitisation,
    EvaluatePulses,
    FreqShiftSanitisation,
    InactivePulseChannelSanitisation,
    InitialPhaseResetSanitisation,
    InstructionGranularitySanitisation,
    InstructionLengthSanitisation,
    LowerSyncsToDelays,
    MeasurePhaseResetSanitisation,
    PhaseOptimisation,
    PostProcessingSanitisation,
    RepeatSanitisation,
    RepeatTranslation,
    ResetsToDelays,
    ReturnSanitisation,
    ScopeSanitisation,
    SquashDelaysOptimisation,
    SynchronizeTask,
)
from qat.middleend.passes.purr.validation import (
    DynamicFrequencyValidation,
    FrequencySetupValidation,
    HardwareConfigValidity,
    ReadoutValidation,
)
from qat.middleend.passes.transform import (
    PydAcquireSanitisation,
    PydBatchedShots,
    PydEndOfTaskResetSanitisation,
    PydEvaluateWaveforms,
    PydFreqShiftSanitisation,
    PydInactivePulseChannelSanitisation,
    PydInitialPhaseResetSanitisation,
    PydInstructionGranularitySanitisation,
    PydInstructionLengthSanitisation,
    PydLowerSyncsToDelays,
    PydMeasurePhaseResetSanitisation,
    PydPhaseOptimisation,
    PydPostProcessingSanitisation,
    PydRepeatSanitisation,
    PydRepeatTranslation,
    PydResetsToDelays,
    PydReturnSanitisation,
    PydScopeSanitisation,
    PydSquashDelaysOptimisation,
    PydSynchronizeTask,
)
from qat.middleend.passes.validation import (
    PydDynamicFrequencyValidation,
    PydFrequencySetupValidation,
    PydHardwareConfigValidity,
    PydReadoutValidation,
)
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.model.target_data import TargetData
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class BaseMiddleend(abc.ABC):
    """
    Base class for a middle end that takes an intermediate representation (IR) :class:`QatIR`
    and alters it based on optimisation and/or validation passes.
    """

    def __init__(self, model: None | QuantumHardwareModel):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = model

    @abc.abstractmethod
    def emit(
        self,
        ir,
        res_mgr: Optional[ResultManager] = None,
        met_mgr: Optional[MetricsManager] = None,
        compiler_config: Optional[CompilerConfig] = None,
    ):
        """
        Converts an IR :class:`QatIR` to an optimised IR.
        :param ir: The intermediate representation.
        :param res_mgr: Collection of analysis results with caching and aggregation
                        capabilities, defaults to None.
        :param met_mgr: Stores useful intermediary metrics that are generated during
                        compilation, defaults to None.
        :param compiler_config: Compiler settings, defaults to None.
        """
        ...


class CustomMiddleend(BaseMiddleend):
    """
    Middle end that uses a custom pipeline to convert the IR to an (optimised) IR.
    """

    def __init__(
        self, model: None | QuantumHardwareModel, pipeline: None | PassManager = None
    ):
        self.pipeline = pipeline
        super().__init__(model=model)

    def emit(
        self,
        ir,
        res_mgr: Optional[ResultManager] = None,
        met_mgr: Optional[MetricsManager] = None,
        compiler_config: Optional[CompilerConfig] = None,
    ):
        """
        Converts an IR :class:`QatIR` to an optimised IR with a custom pipeline.
        :param ir: The intermediate representation.
        :param res_mgr: Collection of analysis results with caching and aggregation
                        capabilities, defaults to None.
        :param met_mgr: Stores useful intermediary metrics that are generated during
                        compilation, defaults to None.
        :param compiler_config: Compiler settings, defaults to None.
        """

        res_mgr = res_mgr if res_mgr is not None else ResultManager()
        met_mgr = met_mgr if met_mgr is not None else MetricsManager()
        compiler_config = (
            compiler_config if compiler_config is not None else CompilerConfig()
        )

        ir = self.pipeline.run(ir, res_mgr, met_mgr, compiler_config=compiler_config)
        return ir


class FallthroughMiddleend(CustomMiddleend):
    """
    A middle end that passes through an input :class:`InstructionBuilder` and does not alter it.
    """

    def __init__(self, model: None = None):
        super().__init__(model=None, pipeline=PassManager())


class DefaultMiddleend(CustomMiddleend):
    """
    Validates the compiler settings against the hardware model, finds calibrations within the IR,
    compresses contiguous :class:`PhaseShift` instructions, checks that the :class:`PostProcessing`
    instructions that follow an acquisition are suitable for the acquisition mode, validates that
    there are no mid-circuit measurements.
    """

    def __init__(
        self,
        model: QuantumHardwareModel,
        target_data: TargetData = TargetData.default(),
    ):
        """
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        :param clock_cycle: The period for a single sequencer clock cycle.
        """
        pipeline = self.build_pass_pipeline(model, target_data)
        self.target_data = target_data
        super().__init__(model=model, pipeline=pipeline)

    @staticmethod
    def build_pass_pipeline(
        model: QuantumHardwareModel, target_data: TargetData = TargetData.default()
    ) -> PassManager:
        """
        Builds the default middle end pass pipeline.
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        :return: A :class:`PassManager` containing a sequence of passes.
        """
        return (
            PassManager()
            | HardwareConfigValidity(model)
            | ActivePulseChannelAnalysis(model)
            | FrequencySetupValidation(model, target_data)
            | DynamicFrequencyValidation(model, target_data)
            # Sanitising input IR to make it complete
            | RepeatSanitisation(model, target_data)
            | ReturnSanitisation()
            | SynchronizeTask()
            # Corrections / optimisations to the IR
            | PostProcessingSanitisation()
            | ReadoutValidation(model)
            | AcquireSanitisation()
            | MeasurePhaseResetSanitisation()
            | InstructionGranularitySanitisation(model, target_data)
            # Preparing for codegen
            | EvaluatePulses()
            | LowerSyncsToDelays()
            | InactivePulseChannelSanitisation()
            | FreqShiftSanitisation(model)
            | InitialPhaseResetSanitisation()
            | PhaseOptimisation()
            | EndOfTaskResetSanitisation()
            | ResetsToDelays(target_data)
            | SquashDelaysOptimisation()
            | InstructionLengthSanitisation(model, target_data)
            | BatchedShots(target_data)
            | ScopeSanitisation()
            | RepeatTranslation(target_data)
        )


class ExperimentalDefaultMiddleend(CustomMiddleend):
    """
    Validates the compiler settings against the hardware model, finds calibrations within the IR,
    compresses contiguous :class:`PhaseShift` instructions, checks that the :class:`PostProcessing`
    instructions that follow an acquisition are suitable for the acquisition mode, validates that
    there are no mid-circuit measurements.
    """

    def __init__(
        self,
        model: QuantumHardwareModel,
        pyd_model: PydHardwareModel,
        target_data: TargetData = TargetData.default(),
    ):
        """
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        :param clock_cycle: The period for a single sequencer clock cycle.
        """
        pipeline = self.build_pass_pipeline(model, pyd_model, target_data)
        self.target_data = target_data
        self.pyd_model = pyd_model
        super().__init__(model=model, pipeline=pipeline)

    @staticmethod
    def build_pass_pipeline(
        legacy_model: QuantumHardwareModel,
        pyd_model: PydHardwareModel,
        target_data: TargetData = TargetData.default(),
    ) -> PassManager:
        """
        Builds the default middle end pass pipeline.
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        :return: A :class:`PassManager` containing a sequence of passes.
        """
        return (
            PassManager()
            | ConvertToPydanticIR(legacy_model, pyd_model)
            | PydHardwareConfigValidity(pyd_model)
            | PydActivePulseChannelAnalysis(pyd_model)
            | PydFrequencySetupValidation(pyd_model, target_data)
            | PydDynamicFrequencyValidation(pyd_model, target_data)
            # Sanitising input IR to make it complete
            | PydRepeatSanitisation(target_data)  # TODO: COMPILER-347
            | PydReturnSanitisation()
            | PydSynchronizeTask()
            | PydPostProcessingSanitisation()
            | PydReadoutValidation()
            | PydAcquireSanitisation()
            | PydMeasurePhaseResetSanitisation(pyd_model)
            | PydInstructionGranularitySanitisation(pyd_model, target_data)
            # Preparing for codegen
            | PydEvaluateWaveforms(pyd_model, target_data)
            | PydLowerSyncsToDelays()
            | PydInactivePulseChannelSanitisation()
            | PydFreqShiftSanitisation(pyd_model)
            | PydInitialPhaseResetSanitisation()
            | PydPhaseOptimisation()
            | PydEndOfTaskResetSanitisation(pyd_model)
            | PydResetsToDelays(pyd_model, target_data)
            | PydSquashDelaysOptimisation()
            | PydInstructionLengthSanitisation(pyd_model, target_data)
            | PydBatchedShots(target_data)
            | PydScopeSanitisation()
            | PydRepeatTranslation(target_data)
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2024 Oxford Quantum Circuits Ltd


from compiler_config.config import TketOptimizations
from pytket._tket.architecture import Architecture
from pytket._tket.predicates import MaxNQubitsPredicate
from pytket.passes import SequencePass
from pytket.qasm import circuit_to_qasm_str
from pytket.qasm.qasm import QASMUnsupportedError

from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.integrations.tket import (
    TketBuilder,
    TketQasmParser,
    _full_stopalize,
    apply_default_transforms,
    check_validity,
    fetch_default_passes,
    optimize_circuit,
)
from qat.purr.utils.logger import get_default_logger
from qat.utils.graphs import get_connected_subgraphs

log = get_default_logger()


def run_pyd_tket_optimizations(qasm_string, opts, hardware: PydHardwareModel) -> str:
    """
    Runs tket-based optimizations and modifications given a Pydantic hardware model.
    Routing will always happen no matter the level.

    Will run optimizations in sections if a full suite fails until a minimal subset of
    passing optimizations is found.
    """

    try:
        tket_builder: TketBuilder = TketQasmParser().parse(TketBuilder(), qasm_string)
        circ = tket_builder.circuit
        log.info(f"Number of gates before tket optimization: {circ.n_gates}")
    except Exception as e:  # Parsing is too fragile, can cause almost any exception.
        log.warning(
            f"Tket failed during QASM parsing with error: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )
        return qasm_string

    couplings = []
    coupling_qualities = []
    optimizations_failed = False
    architecture = None

    # Without default remapping pass multi-qubit gates don't get moved around, so
    # trying to apply them to a limited subset of qubits provides no value.
    logical_qubit_map = {q_i: i for i, q_i in enumerate(hardware.qubits.keys())}

    for q, coupled_qs in hardware.logical_connectivity.items():
        logical_q = logical_qubit_map[q]
        for coupled_q in coupled_qs:
            logical_coupled_q = logical_qubit_map[coupled_q]

            couplings.append((logical_q, logical_coupled_q))
            coupling_qualities.append(hardware.logical_connectivity_quality[(q, coupled_q)])

    if TketOptimizations.DefaultMappingPass not in opts:
        architecture = Architecture(couplings)
        optimizations_failed = not optimize_circuit(circ, architecture, opts)
    else:
        sorted_coupling_qualities = sorted(coupling_qualities, reverse=True)
        for quality_level in sorted_coupling_qualities:
            filtered_couplings = [
                couplings[i]
                for i, quality in enumerate(coupling_qualities)
                if quality >= quality_level
            ]

            subgraphs_nodes, subgraphs_edges = get_connected_subgraphs(filtered_couplings)
            for nodes, edges in zip(subgraphs_nodes, subgraphs_edges):
                if circ.n_qubits <= len(nodes):
                    architecture = Architecture(edges)
                    optimizations_failed = not optimize_circuit(circ, architecture, opts)
                    if not optimizations_failed:
                        break
                else:
                    optimizations_failed = True
            if not optimizations_failed:
                break

    # If our optimizations failed but we want the mapping pass, apply that by itself.
    if optimizations_failed:
        if architecture is None:
            raise ValueError(
                "Unable to resolve hardware instance for fall-back optimizations."
            )

        delay_failed = False
        try:
            # DelayMeasure throws on failure, and we want to raise our own errors for
            # this.
            SequencePass(fetch_default_passes(architecture, opts)).apply(circ)
        except RuntimeError:
            delay_failed = True

        # If the delay fails, try with a more limited subset.
        if delay_failed:
            try:
                # Tket just throws an exception if the list is none, so skip if that's
                # the case.
                default_passes = fetch_default_passes(architecture, opts, add_delay=False)
                if len(default_passes) > 0:
                    SequencePass(default_passes).apply(circ)
            except RuntimeError as e:
                message = str(e)
                if MaxNQubitsPredicate.__name__ in message:
                    raise ValueError(
                        f"Circuit uses {len(circ.qubits)} qubits, "
                        f"only {len(architecture.nodes)} available."
                    ) from e

                raise e

        apply_default_transforms(circ, architecture, opts)
        check_validity(circ, architecture)

    try:
        qasm_string = circuit_to_qasm_str(circ)
        log.info(f"Number of gates after tket optimization: {circ.n_gates}")
    except (QASMUnsupportedError, RuntimeError) as e:
        log.warning(
            f"Error generating QASM from Tket circuit: {_full_stopalize(e)}. "
            "Skipping this optimization pass."
        )

    # TODO: Return result object with more information about compilation/errors
    return qasm_string



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import itertools

import numpy as np
from more_itertools import partition

from qat.purr.backends.qblox.pass_base import QatIR, TransformPass
from qat.purr.backends.qblox.result_base import ResultManager
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    EndRepeat,
    EndSweep,
    Repeat,
    Return,
    Sweep,
)


class ScopeSanitisation(TransformPass):
    def run(self, ir: QatIR, res_mgr: ResultManager, *args, **kwargs):
        """
        Bubbles up all sweeps and repeats to the beginning of the list.
        Adds delimiter instructions to the repeats and sweeps signifying the end of their scopes.

        Intended for legacy existing builders and the relative order of instructions guarantees backwards
        compatibility.
        """

        builder = ir.value
        if not isinstance(builder, InstructionBuilder):
            raise ValueError(f"Expected InstructionBuilder, got {type(builder)}")

        tail, head = partition(
            lambda inst: isinstance(inst, (Sweep, Repeat)), builder.instructions
        )
        tail, head = list(tail), list(head)

        delimiters = [
            EndSweep() if isinstance(inst, Sweep) else EndRepeat() for inst in head
        ]

        builder.instructions = head + tail + delimiters[::-1]


class RepeatSanitisation(TransformPass):
    def __init__(self, model: QuantumHardwareModel = None):
        self.model = model

    def run(self, ir: QatIR, res_mgr: ResultManager, *args, **kwargs):
        """
        Fixes repeat instructions if any with default values from the HW model
        Adds a repeat instructions if none is found
        """

        builder = ir.value
        if not isinstance(builder, InstructionBuilder):
            raise ValueError(f"Expected InstructionBuilder, got {type(builder)}")

        model = self.model

        repeats = [inst for inst in builder.instructions if isinstance(inst, Repeat)]
        if repeats:
            for rep in repeats:
                if rep.repeat_count is None:
                    rep.repeat_count = model.default_repeat_count
                if rep.repetition_period is None:
                    rep.repetition_period = model.default_repetition_period
        else:
            builder.repeat(model.default_repeat_count, model.default_repetition_period)


class ReturnSanitisation(TransformPass):
    def run(self, ir: QatIR, res_mgr: ResultManager, *args, **kwargs):
        builder = ir.value
        if not isinstance(builder, InstructionBuilder):
            raise ValueError(f"Expected InstructionBuilder, got {type(builder)}")

        returns = [inst for inst in builder.instructions if isinstance(inst, Return)]
        acquires = [inst for inst in builder.instructions if isinstance(inst, Acquire)]

        if returns:
            unique_variables = set(itertools.chain(*[ret.variables for ret in returns]))
            for ret in returns:
                builder.instructions.remove(ret)
        else:
            # If we don't have an explicit return, imply all results.
            unique_variables = set(acq.output_variable for acq in acquires)

        builder.returns(list(unique_variables))


class DesugaringPass(TransformPass):
    """
    Transforms syntactic sugars and implicit effects. Not that it applies here, but a classical
    example of a sugar syntax is the ternary expression "cond ? val1 : val2" which is nothing more
    than an if else in disguise.

    The goal of this pass is to desugar any syntactic and semantic sugars. One of these sugars
    is iteration constructs such as Sweep and Repeat.
    """

    def run(self, ir: QatIR, res_mgr: ResultManager, *args, **kwargs):
        builder = ir.value
        if not isinstance(builder, InstructionBuilder):
            raise ValueError(f"Expected InstructionBuilder, got {type(builder)}")

        for inst in builder.instructions:
            if isinstance(inst, Sweep):
                count = len(next(iter(inst.variables.values())))
                iter_name = f"sweep_{hash(inst)}"
                inst.variables[iter_name] = np.linspace(1, count, count, dtype=int).tolist()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import re
from collections.abc import Iterable
from copy import deepcopy
from functools import cached_property
from pydoc import locate
from typing import Annotated, Any, Type, TypeVar, Union, get_args, get_origin

import numpy as np
from frozendict import frozendict
from numpydantic import NDArray, Shape
from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    PlainSerializer,
    PlainValidator,
    PrivateAttr,
    RootModel,
    computed_field,
    model_validator,
)
from pydantic._internal._model_construction import ModelMetaclass
from pydantic_core import core_schema

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class NoExtraFieldsModel(BaseModel):
    """
    A Pydantic `BaseModel` with the extra constraints:
        #. Assignment of fields after initialisation is checked again.
        #. Extra fields given to the model are not ignored (default behaviour in `BaseModel`),
          but raise an error now.
    """

    model_config = ConfigDict(
        validate_assignment=True,
        use_enum_values=False,
        extra="forbid",
        ser_json_inf_nan="constants",
    )

    def __str__(self):
        return self.__repr__()


class NoExtraFieldsFrozenModel(NoExtraFieldsModel):
    """
    A Pydantic `BaseModel` with the extra constraints:
        #. Assignment of fields after initialisation is checked again.
        #. Extra fields given to the model are not ignored (default behaviour in `BaseModel`),
          but raise an error now.
        # All fields are frozen upon instantiation.
    """

    model_config = ConfigDict(frozen=True)


class AllowExtraFieldsModel(BaseModel):
    """
    A Pydantic `BaseModel` with the extra constraints:
        #. Assignment of fields after initialisation is checked again.
        #. Extra fields given to the model are ignored (default behaviour in `BaseModel`).
    """

    model_config = ConfigDict(
        validate_assignment=True, use_enum_values=False, extra="ignore"
    )

    def __str__(self):
        return self.__repr__()


class RehydratableModel(BaseModel):
    @computed_field
    @cached_property
    def object_type(self) -> str:
        """
        Returns the type of the object, which is the class name.
        """
        return self.__class__.__module__ + "." + self.__class__.__name__

    @classmethod
    def _rehydrate_object(cls, data):
        type_str = data.get("object_type", None)
        if type_str is None:
            cls_type = cls
        else:
            # Locate the class using the type string.
            cls_type = locate(type_str)
        if cls_type is None:
            raise ValueError(
                f"Could not locate class for type string '{type_str}'. Ensure it is a valid class path."
            )
        # Create an instance of the located class with the provided data.
        if not issubclass(cls_type, cls):
            raise TypeError(
                f"Located class '{cls_type.__name__}' is not a subclass of '{cls.__name__}'."
            )
        return cls_type(**data)


# This base file is used to implement classes/methods common to all hardware.


def validate_non_negative(value: int):
    if not isinstance(value, int) or value < 0:
        raise ValueError(f"Given value {value} must be an int and >=0.")
    return value


NonNegativeInt = Annotated[
    int,
    AfterValidator(validate_non_negative),
]


def validate_calibratable_positive_float(value: CalibratablePositiveFloat):
    if not np.isnan(value) and value < 0.0:
        raise ValueError(f"Given value {value} must be >=0.")
    return value


CalibratablePositiveFloat = Annotated[
    float,
    AfterValidator(validate_calibratable_positive_float),
]


def validate_calibratable_unit_interval(value: CalibratableUnitInterval):
    if not np.isnan(value):
        if value < 0.0 or value > 1.0:
            raise ValueError(f"Given value {value} must be in the interval [0, 1].")
    return value


# A float in the unit interval [0, 1].
CalibratableUnitInterval = Annotated[
    float,
    AfterValidator(validate_calibratable_unit_interval),
]


def validate_calibratable_unit_interval_array(array: CalibratableUnitInterval2x2Array):
    if np.any(array > 1) or np.any(array < 0):
        raise ValueError("Given array elements must be in the interval [0, 1].")
    return array


CalibratableUnitInterval2x2Array = Annotated[
    NDArray[Shape["2, 2"], float], AfterValidator(validate_calibratable_unit_interval_array)
]


def validate_qubit_coupling(value: QubitCoupling):
    if isinstance(value, str):
        return tuple(map(int, re.findall(r"\d+", value)))
    elif isinstance(value, tuple):
        return value
    else:
        raise TypeError(
            "Invalid type for `QubitCoupling`. Please provide a `str` or `tuple`."
        )


# A qubit coupling represented as (q_i: int, q_j: int).
QubitCoupling = Annotated[
    tuple[NonNegativeInt, NonNegativeInt],
    BeforeValidator(validate_qubit_coupling),
]

VALIDATORS = Union[BeforeValidator, AfterValidator, PlainValidator]


def validate_value(value: V, value_type: Type[V]):
    # Get the validator function if it's an annotated type.
    f_validate = None
    if get_origin(value_type) is Annotated:
        value_type, *metadata = get_args(value_type)
        for item in metadata:
            if isinstance(item, VALIDATORS):
                f_validate = item.func

    if get_origin(value_type) is Union:
        allowed_types = get_args(value_type)
    else:
        allowed_types = (value_type,)

    # Validate if type of value == `V`.
    if not isinstance(value, allowed_types):
        raise TypeError(
            f"Cannot add value {value} of type '{type(value)} to container of type {value_type}'."
        )

    # Validate extra constraints on the value, provided via the annotation.
    if f_validate:
        value = f_validate(value)


K = TypeVar("GeneralKey")
V = TypeVar("GeneralValue")


class PydListBase(RootModel[list[V]]):
    root: list[V] = Field(default_factory=list)
    _value_type: type = PrivateAttr()

    @model_validator(mode="after")
    def container_type(self):
        """
        Validate the type of elements in the list.
        """
        value_type = get_args(self.__class__.model_fields["root"].annotation)[0]
        self._value_type = value_type
        return self

    def __iter__(self):
        return iter(self.root)

    def __eq__(self, other: PydListBase | list):
        if isinstance(other, PydListBase):
            return self.root.__eq__(other.root)
        elif isinstance(other, list):
            return self.root.__eq__(other)
        return False

    def __getitem__(self, i: int):
        return self.root.__getitem__(i)

    def __len__(self):
        return self.root.__len__()

    def __repr__(self):
        return self.root.__repr__()

    def __str__(self):
        return self.root.__str__()

    def __deepcopy__(self, memo):
        copied_root = deepcopy(self.root, memo)
        copied_instance = self.__class__(root=copied_root)
        memo[id(self)] = copied_instance
        return copied_instance


class ValidatedList(PydListBase):
    """
    A list object that validates the input appended/extended after instantiation.
    This way, we are sure that the elements in a list are only of a certain type.
    (FYI: Pydantic containers only validate upon instantiation.)
    """

    @model_validator(mode="after")
    def validate_root(self):
        for value in self.root:
            validate_value(value, self._value_type)
        return self

    def append(self, value: V):
        validate_value(value, self._value_type)
        self.root.append(value)

    def extend(self, values: Iterable[V]):
        for value in values:
            self.append(value)

    def remove(self, value: V):
        self.root.remove(value)


def _validate_set(value: float | int | str | Iterable | None):
    if isinstance(value, (float, int, str)):
        value = {value}
    elif isinstance(value, (list, ValidatedList, tuple)):
        value = set(value)
    return value


class PydSetBase(RootModel[set[V]]):
    root: set[V] = Field(default_factory=set)
    _value_type: type = PrivateAttr()

    @model_validator(mode="after")
    def container_type(self):
        """
        Validate the types of keys and values in the dictionary.
        """
        value_type = get_args(self.__class__.model_fields["root"].annotation)[0]
        self._value_type = value_type
        return self

    def __iter__(self):
        return iter(self.root)

    def __eq__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.__eq__(other.root)
        elif isinstance(other, set):
            return self.root.__eq__(other)
        elif isinstance(other, (list, tuple)):
            return self.root.__eq__(set(other))
        return False

    def __le__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.issubset(other.root)
        elif isinstance(other, set):
            return self.root.issubset(other)
        else:
            raise NotImplementedError(f"Unsupported processing for {self} and {other}.")

    def __sub__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.__sub__(other.root)
        elif isinstance(other, set):
            return self.root.__sub__(other)
        else:
            raise NotImplementedError(f"Unsupported processing for {self} and {other}.")

    def __and__(self, other: PydSetBase):
        if isinstance(other, PydSetBase):
            return self.root.__and__(other.root)
        elif isinstance(other, set):
            return self.root.__and__(other)
        else:
            raise NotImplementedError(f"Unsupported processing for {self} and {other}.")

    def __len__(self):
        return self.root.__len__()

    def __repr__(self):
        return self.root.__repr__()

    def __str__(self):
        return self.root.__str__()

    def __deepcopy__(self, memo):
        copied_root = deepcopy(self.root, memo)
        copied_instance = self.__class__(root=copied_root)
        memo[id(self)] = copied_instance
        return copied_instance


class FrozenSet(PydSetBase):
    """
    A Pydantic set that is immutable after instantiation.
    """

    root: frozenset[V] = Field(default_factory=frozenset)


class ValidatedSet(PydSetBase):
    """
    A set object that validates the input added after instantiation.
    This way, we are sure that the elements in a set are only of a certain type.
    (FYI: Pydantic containers only validate upon instantiation.)
    """

    @model_validator(mode="after")
    def validate_root(self):
        for value in self.root:
            validate_value(value, self._value_type)
        return self

    def add(self, value: V):
        validate_value(value, self._value_type)
        self.root.add(value)

    def discard(self, value):
        self.root.discard(value)

    def pop(self):
        self.root.pop()

    def remove(self, value):
        self.root.remove(value)

    def update(self, *sets):
        for s in sets:
            for value in s:
                self.add(value)


class PydDictBase(RootModel[dict[K, V]]):
    root: dict[K, V] = Field(default_factory=dict)
    _key_type: type = PrivateAttr()
    _value_type: type = PrivateAttr()

    @model_validator(mode="after")
    def container_types(self):
        """
        Validate the types of keys and values in the dictionary.
        """
        key_type, value_type = get_args(self.__class__.model_fields["root"].annotation)
        self._key_type = key_type
        self._value_type = value_type
        return self

    def get(self, key, default=None):
        return self.root.get(key, default)

    def keys(self):
        return self.root.keys()

    def items(self):
        return self.root.items()

    def values(self):
        return self.root.values()

    def __eq__(self, other: PydDictBase):
        if isinstance(other, PydDictBase):
            return self.root.__eq__(other.root)
        elif isinstance(other, dict):
            return self.root.__eq__(other)
        return False

    def __getitem__(self, key):
        return self.root.get(key, None)

    def pop(self, key: Any, *args, **kwargs):
        return self.root.pop(key, *args, **kwargs)

    def __iter__(self):
        return iter(self.root)

    def __len__(self):
        return self.root.__len__()

    def __repr__(self):
        return self.root.__repr__()

    def __str__(self):
        return self.root.__str__()

    def __deepcopy__(self, memo):
        result = {}
        for k, v in self.root.items():
            result[deepcopy(k)] = deepcopy(v)

        result = self.__class__(result)
        memo[id(self)] = result
        return result


class _PydanticFrozenDictAnnotation:
    """
    Helper class since Pydantic `V2` does only offer support for `frozenset`, not `frozendict`.
    """

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        def validate_from_dict(d: dict | frozendict) -> frozendict:
            return frozendict(d)

        k, v = get_args(source_type)
        frozendict_schema = core_schema.chain_schema(
            [
                handler.generate_schema(dict[k, v]),
                core_schema.no_info_plain_validator_function(validate_from_dict),
                core_schema.is_instance_schema(frozendict),
            ]
        )
        return core_schema.json_or_python_schema(
            json_schema=frozendict_schema,
            python_schema=frozendict_schema,
            serialization=core_schema.plain_serializer_function_ser_schema(dict),
        )


pyd_frozendict = Annotated[frozendict[K, V], _PydanticFrozenDictAnnotation]


class FrozenDict(PydDictBase):
    """
    A Pydantic dict that is immutable after instantiation.
    """

    root: pyd_frozendict[K, V] = Field(default_factory=frozendict)


class ValidatedDict(PydDictBase):
    """
    A dict object that validates the input added after instantiation.
    This way, we are sure that the elements in a dict are only of a certain type.
    (FYI: Pydantic containers only validate upon instantiation.)
    """

    @model_validator(mode="after")
    def validate_root(self):
        for key, value in self.root.items():
            validate_value(key, self._key_type)
            validate_value(value, self._value_type)
        return self

    def update(self, data: dict[K, V]):
        for key, value in data.items():
            self.__setitem__(key, value)

    def __setitem__(self, key: K, value: V):
        validate_value(key, self._key_type)
        validate_value(value, self._value_type)
        self.root[key] = value


QubitId = NonNegativeInt


def validate_waveform_type(value: BaseModel):
    if isinstance(value, str):
        return locate("qat.ir.waveforms." + value)
    elif isinstance(value, ModelMetaclass):
        return value
    else:
        raise TypeError(
            "Invalid type for `WaveformType`. Please provide a `str` or `BaseModel`."
        )


WaveformType = Annotated[
    TypeVar("Waveform"),
    PlainSerializer(lambda wf: wf.__name__, when_used="json-unless-none"),
    BeforeValidator(validate_waveform_type),
]


def find_all_subclasses(cls: Type) -> list[Type]:
    """
    Recursively finds nested subclasses of a class.
    """
    subclasses = cls.__subclasses__()
    for subclass in subclasses:
        subclasses.extend(find_all_subclasses(subclass))
    return subclasses


# Efficient serializing of numeric numpy arrays
def _list_serializer(lst):
    """Lists of complex numbers can be expensive to serialise: by serializing type
    information and its value as a hex, we can have more performant serialization."""
    lst = np.asarray(lst)
    return {"dtype": lst.dtype.name, "shape": lst.shape, "value": lst.tobytes().hex()}


def _list_validator(lst, ty: type):
    """Reverts the hex value and type information into a numpy array."""
    if isinstance(lst, dict):
        arr = np.frombuffer(bytearray.fromhex(lst["value"]), dtype=np.dtype(lst["dtype"]))
        return arr.reshape(lst["shape"])
    if isinstance(lst, list):
        return np.asarray(lst, dtype=ty)
    return lst


def get_annotated_array(ty: type):
    """Creates an annotated type for numeric lists with Pydantic serializers and validators
    for efficient serialization."""
    return Annotated[
        # TODO: Investigate linting issue with Shape["* x"]
        NDArray[Shape["* x"], ty],  # noqa: F722
        BeforeValidator(lambda x: _list_validator(x, ty)),
        PlainSerializer(_list_serializer),
    ]


IntNDArray = get_annotated_array(int)
FloatNDArray = get_annotated_array(float)
ComplexNDArray = get_annotated_array(complex)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import warnings
from typing import Optional

import numpy as np

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.executables import Executable
from qat.runtime import BaseRuntime, ResultsAggregator
from qat.runtime.connection import ConnectionMode


class SimpleRuntime(BaseRuntime):
    """A Runtime for the complete execution of packages without sweeps.

    The :class:`SimpleRuntime` handles the complete execution for simple programs that are
    free from sweeps (with exceptions of sweeps that have been lowered to the hardware).
    This includes batching of shots, executing the program on the target machine, and any software
    post-processing that cannot be achieved in the target machine. The runtime must be provided
    with a :class:`NativeEngine` that is capable of executing the desired programs.
    """

    def execute(
        self,
        package: Executable,
        res_mgr: Optional[ResultManager] = None,
        met_mgr: Optional[MetricsManager] = None,
        **kwargs,
    ):
        """Fully execute a package against the hardware with batching of shots and results
        post-processing.

        :param package: The executable program.
        :param res_mgr: Optionally provide a results manager to save pass information.
        :param met_mgr: Optionally provide a metric manager to save pass information.
        :returns: Execution results.
        """

        if res_mgr is None:
            res_mgr = ResultManager()
        if met_mgr is None:
            met_mgr = MetricsManager()

        self.validate_max_shots(package.shots)
        if package.shots == 0:
            warnings.warn("Tried to execute a package with zero shots.")
            return {acquire.output_variable: np.array([]) for acquire in package.acquires}

        number_of_batches = self.number_of_batches(package.shots, package.compiled_shots)
        self.connect_engine(ConnectionMode.CONNECT_BEFORE_EXECUTE)
        aggregator = ResultsAggregator(package.acquires)
        for _ in range(number_of_batches):
            batch_results = self.engine.execute(package)
            aggregator.update(batch_results)
        self.disconnect_engine(ConnectionMode.DISCONNECT_AFTER_EXECUTE)

        acquisitions = aggregator.results(package.shots)
        return self.results_pipeline.run(
            acquisitions, res_mgr, met_mgr, package=package, **kwargs
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.core.pass_base import PassManager
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.runtime.passes.analysis import IndexMappingAnalysis as PydIndexMappingAnalysis
from qat.runtime.passes.purr.analysis import IndexMappingAnalysis
from qat.runtime.passes.transform import (
    AssignResultsTransform,
    ErrorMitigation,
    InlineResultsProcessingTransform,
    PostProcessingTransform,
    ResultTransform,
)


def get_default_results_pipeline(model: QuantumHardwareModel) -> PassManager:
    """Factory for creating the default results pipeline.

    :param model: The quantum hardware model to use for the pipeline.
    """

    return (
        PassManager()
        | PostProcessingTransform()
        | InlineResultsProcessingTransform()
        | AssignResultsTransform()
        | ResultTransform()
        | IndexMappingAnalysis(model)
        | ErrorMitigation(model)
    )


def get_experimental_results_pipeline(
    model: QuantumHardwareModel, pyd_model: PydHardwareModel
) -> PassManager:
    """Factory for creating the experimental results pipeline.

    This pipeline includes additional experimental features.

    :param model: The quantum hardware model to use for the pipeline.
    :param pyd_model: The quantum hardware model converted to Pydantic.
    """

    return (
        PassManager()
        | PostProcessingTransform()
        | InlineResultsProcessingTransform()
        | AssignResultsTransform()
        | ResultTransform()
        | PydIndexMappingAnalysis(pyd_model)
        | ErrorMitigation(pyd_model)
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.frontend.auto import AutoFrontend as AutoFrontend
from qat.frontend.auto import AutoFrontendWithFlattenedIR as AutoFrontendWithFlattenedIR
from qat.frontend.base import BaseFrontend as BaseFrontend
from qat.frontend.custom import CustomFrontend as CustomFrontend
from qat.frontend.fallthrough import FallthroughFrontend as FallthroughFrontend
from qat.frontend.qasm import Qasm2Frontend as Qasm2Frontend
from qat.frontend.qasm import Qasm3Frontend as Qasm3Frontend
from qat.frontend.qir import QIRFrontend as QIRFrontend

DefaultFrontend = AutoFrontend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Oxford Quantum Circuits Ltd
from collections import defaultdict
from functools import singledispatchmethod

import numpy as np
from compiler_config.config import CompilerConfig, ErrorMitigationConfig, ResultsFormatting

from qat.core.config.configure import get_config
from qat.core.pass_base import ValidationPass
from qat.core.result_base import ResultManager
from qat.ir.instruction_builder import InstructionBuilder
from qat.ir.instructions import FrequencySet, FrequencyShift, Instruction, Repeat, Return
from qat.ir.measure import (
    Acquire,
    AcquireMode,
    PostProcessing,
    ProcessAxis,
    Pulse,
)
from qat.middleend.passes.analysis import ActivePulseChannelResults
from qat.model.device import PhysicalChannel
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.target_data import TargetData
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class DynamicFrequencyValidation(ValidationPass):
    """Validates the setting or shifting frequencies does not move the intermediate
    frequency of a pulse channel outside the allowed limits."""

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        :param target_data: Target-related information.
        """
        self.model = model
        self._is_resonator = self._create_resonator_map(model)

        # Extract limits from the target data.
        self.qubit_if_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_if_freq_min,
            target_data.QUBIT_DATA.pulse_channel_if_freq_max,
        )
        self.resonator_if_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_if_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_if_freq_max,
        )

    @staticmethod
    def _create_resonator_map(model: PhysicalHardwareModel) -> dict[str, bool]:
        """Creates a map to lookup if physical channels belong to resonators.."""
        res_map = {}
        for qubit in model.qubits.values():
            res_map[qubit.physical_channel.uuid] = False
            res_map[qubit.resonator.physical_channel.uuid] = True
        return res_map

    def _validate_frequency_shifts(self, physical_channel_id: str, ifs: list[float]):
        """Validates that the frequency shifts do not exceed the allowed limits."""

        if_limits = (
            self.qubit_if_freq_limits
            if not self._is_resonator[physical_channel_id]
            else self.resonator_if_freq_limits
        )

        return [
            if_value
            for if_value in ifs
            if not (if_limits[0] <= np.abs(if_value) <= if_limits[1])
        ]

    @singledispatchmethod
    def _calculate_frequency(
        self,
        instruction: Instruction,
        ifs: dict[str, list[float]],
        physical_channel_ids: dict[str, str],
    ):
        pass

    @_calculate_frequency.register(FrequencyShift)
    def _(
        self,
        instruction: FrequencyShift,
        ifs: dict[str, list[float]],
        physical_channel_ids: dict[str, str],
    ):
        frequencies = ifs[instruction.target]
        physical_channel = self.model.physical_channel_for_pulse_channel_id(
            instruction.target
        )
        freq = (
            frequencies[-1]
            if len(frequencies) > 0
            else (
                self.model.pulse_channel_with_id(instruction.target).frequency
                - physical_channel.baseband.frequency
            )
        )
        frequencies.append(freq + instruction.frequency)
        physical_channel_ids[instruction.target] = physical_channel.uuid

    @_calculate_frequency.register(FrequencySet)
    def _(
        self,
        instruction: FrequencySet,
        ifs: dict[str, list[float]],
        physical_channel_ids: dict[str, str],
    ):
        physical_channel = self.model.physical_channel_for_pulse_channel_id(
            instruction.target
        )
        ifs[instruction.target].append(
            instruction.frequency - physical_channel.baseband.frequency
        )
        physical_channel_ids[instruction.target] = physical_channel.uuid

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        ifs = defaultdict(list)
        physical_channel_ids = defaultdict(str)
        for instruction in ir.instructions:
            self._calculate_frequency(instruction, ifs, physical_channel_ids)

        violations = []
        for pulse_channel_id, if_values in ifs.items():
            physical_channel_id = physical_channel_ids[pulse_channel_id]
            if if_violations := self._validate_frequency_shifts(
                physical_channel_id, if_values
            ):
                pulse_channel = self.model.pulse_channel_with_id(pulse_channel_id)
                device = self.model.device_for_pulse_channel_id(pulse_channel_id)
                device_name = "Qubit "
                if self._is_resonator[physical_channel_id]:
                    device_name = "Resonator "
                    device = next(
                        filter(lambda qubit: qubit.resonator is device, self.model.qubits)
                    )
                device_name += str(self.model.index_of_qubit(device))
                violations.append(
                    f"The IF of {device_name} {pulse_channel.pulse_type} pulse channel is set or "
                    f"shifted to values {if_violations} that exceed the allowed limits."
                )

        if len(violations) > 0:
            raise ValueError(
                "Dynamic frequency validation failed with the following violations:\n"
                + "\n".join(violations)
            )
        return ir


class NoMidCircuitMeasurementValidation(ValidationPass):
    """
    Validates that there are no mid-circuit measurements by checking that no qubit
    has an acquire instruction that is later followed by a pulse instruction.
    """

    def __init__(
        self,
        model: PhysicalHardwareModel,
        no_mid_circuit_measurement: bool | None = None,
        *args,
        **kwargs,
    ):
        """
        :param model: The hardware model.
        :param no_mid_circuit_measurement: Whether mid-circuit measurements are allowed.
            If None, uses the default from the QatConfig.
        """
        self.model = model
        self.no_mid_circuit_measurement = (
            no_mid_circuit_measurement
            if no_mid_circuit_measurement is not None
            else get_config().INSTRUCTION_VALIDATION.NO_MID_CIRCUIT_MEASUREMENT
        )

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The intermediate representation (IR) :class:`InstructionBuilder`.
        """
        consumed_acquire_pc: set[str] = set()

        if not self.no_mid_circuit_measurement:
            return ir

        drive_acq_pc_map = {
            qubit.drive_pulse_channel.uuid: qubit.acquire_pulse_channel.uuid
            for qubit in self.model.qubits.values()
        }

        for instr in ir:
            if isinstance(instr, Acquire):
                consumed_acquire_pc.add(instr.target)

            # Check if we have a measure in the middle of the circuit somewhere.
            elif isinstance(instr, Pulse):
                acq_pc = drive_acq_pc_map.get(instr.target, None)

                if acq_pc and acq_pc in consumed_acquire_pc:
                    raise ValueError(
                        "Mid-circuit measurements currently unable to be used."
                    )
        return ir


class ReadoutValidation(ValidationPass):
    """Validates that the post-processing instructions do not have an invalid sequence.

    Extracted from :meth:`qat.purr.backends.live.LiveDeviceEngine.validate`.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        acquire_modes = {}
        for inst in ir:
            if isinstance(inst, Acquire):
                acquire_modes[inst.output_variable] = inst.mode
            if isinstance(inst, PostProcessing):
                acquire_mode = acquire_modes.get(inst.output_variable, None)
                self._post_processing_options_handling(inst, acquire_mode)
        return ir

    @staticmethod
    def _post_processing_options_handling(
        inst: PostProcessing, acquire_mode: AcquireMode | None
    ):
        if acquire_mode == AcquireMode.SCOPE and ProcessAxis.SEQUENCE in inst.axes:
            raise ValueError(
                "Invalid post-processing! Post-processing over SEQUENCE is "
                "not possible after the result is returned from hardware "
                "in SCOPE mode!"
            )
        elif acquire_mode == AcquireMode.INTEGRATOR and ProcessAxis.TIME in inst.axes:
            raise ValueError(
                "Invalid post-processing! Post-processing over TIME is not "
                "possible after the result is returned from hardware in "
                "INTEGRATOR mode!"
            )
        elif acquire_mode == AcquireMode.RAW:
            raise ValueError(
                "Invalid acquire mode! The live hardware doesn't support RAW acquire mode!"
            )
        elif acquire_mode is None:
            raise ValueError(
                f"No AcquireMode found with output variable {inst.output_variable},"
                f"ensure PostProcessing output_variable matches an Acquire output_variable with a"
                f"valid AcquireMode selected."
            )


class HardwareConfigValidity(ValidationPass):
    """Validates the :class:`CompilerConfig` against the hardware model."""

    def __init__(self, hardware_model: PhysicalHardwareModel, max_shots: int | None = None):
        """Instantiate the pass with a hardware model.

        :param hardware_model: The hardware model.
        :param max_shots: The maximum number of shots allowed for a single task.
                If None, uses the default from the QatConfig.
        """
        self.hardware_model = hardware_model
        self.max_shots = (
            max_shots if max_shots is not None else get_config().MAX_REPEATS_LIMIT
        )

    def run(
        self,
        ir: InstructionBuilder,
        *args,
        compiler_config: CompilerConfig,
        **kwargs,
    ):
        self._validate_shots(compiler_config)
        self._validate_error_mitigation(self.hardware_model, compiler_config)
        return ir

    def _validate_shots(self, compiler_config: CompilerConfig):
        shots = compiler_config.repeats
        if shots is not None and shots > self.max_shots:
            raise ValueError(
                f"Number of shots in compiler config {compiler_config.repeats} exceeds max "
                f"number of shots {self.max_shots}."
            )

    def _validate_error_mitigation(
        self, hardware_model: PhysicalHardwareModel, compiler_config: CompilerConfig
    ):
        if (
            compiler_config.error_mitigation
            and compiler_config.error_mitigation != ErrorMitigationConfig.Empty
        ):
            if not hardware_model.error_mitigation.is_enabled:
                raise ValueError("Error mitigation not calibrated on this hardware model.")

            if ResultsFormatting.BinaryCount not in compiler_config.results_format:
                raise ValueError(
                    "BinaryCount format required for readout error mitigation."
                )


class FrequencySetupValidation(ValidationPass):
    """Validates the baseband frequencies and intermediate frequencies of pulse channels
    against the target data.
    """

    def __init__(self, model: PhysicalHardwareModel, target_data: TargetData):
        """Instantiate the pass with a hardware model.

        :param model: The hardware model.
        :param target_data: Target-related information.
        """

        # Extract limits from the target data.
        self.qubit_lo_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_lo_freq_min,
            target_data.QUBIT_DATA.pulse_channel_lo_freq_max,
        )
        self.resonator_lo_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_lo_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_lo_freq_max,
        )
        self.qubit_if_freq_limits = (
            target_data.QUBIT_DATA.pulse_channel_if_freq_min,
            target_data.QUBIT_DATA.pulse_channel_if_freq_max,
        )
        self.resonator_if_freq_limits = (
            target_data.RESONATOR_DATA.pulse_channel_if_freq_min,
            target_data.RESONATOR_DATA.pulse_channel_if_freq_max,
        )

        self._pulse_channel_data = self._get_pulse_channel_data(model)
        self._physical_channel_data = self._get_physical_channel_data(model)

    def _get_pulse_channel_data(self, model: PhysicalHardwareModel) -> dict[str, dict]:
        """Extracts pulse channel data from the model."""
        pulse_channel_data = {}

        for qubit in model.qubits.values():
            pulse_channel_data = self._populate_pulse_channel_dict(
                pulse_channel_data, qubit.all_pulse_channels, qubit.physical_channel, False
            )
            pulse_channel_data = self._populate_pulse_channel_dict(
                pulse_channel_data,
                qubit.resonator.all_pulse_channels,
                qubit.resonator.physical_channel,
                True,
            )
        return pulse_channel_data

    def _populate_pulse_channel_dict(
        self,
        pulse_channel_data: dict,
        pulse_channels: list[Pulse],
        physical_channel: PhysicalChannel,
        is_resonator: bool,
    ) -> dict:
        for pulse_channel in pulse_channels:
            if_freq = pulse_channel.frequency - physical_channel.baseband.frequency
            pulse_channel_data[pulse_channel.uuid] = {
                "physical_channel": physical_channel.uuid,
                "is_valid": self._validate_pulse_channel_if(if_freq, is_resonator),
                "is_resonator": is_resonator,
                "if_frequency": if_freq,
                "pulse_type": pulse_channel.pulse_type,
            }
        return pulse_channel_data

    def _get_physical_channel_data(self, model: PhysicalHardwareModel) -> dict[str, dict]:
        """Extracts physical channel data from the model."""
        physical_channel_data = {}
        for q_index, qubit in model.qubits.items():
            physical_channel = qubit.physical_channel
            physical_channel_data[physical_channel.uuid] = (
                self._populate_physical_channel_data(physical_channel, False, q_index)
            )

            physical_channel = qubit.resonator.physical_channel
            physical_channel_data[physical_channel.uuid] = (
                self._populate_physical_channel_data(physical_channel, True, q_index)
            )

        return physical_channel_data

    def _populate_physical_channel_data(
        self, physical_channel: PhysicalChannel, is_resonator: bool, qubit_index: int
    ) -> dict:
        physical_channel_data = {
            "baseband_frequency": physical_channel.baseband.frequency,
            "is_resonator": is_resonator,
            "is_valid": self._validate_baseband_frequency(
                physical_channel.baseband.frequency, is_resonator
            ),
            "index": qubit_index,
        }
        return physical_channel_data

    def _validate_baseband_frequency(self, frequency, is_resonator: bool) -> bool:
        """Validates that the frequencies in the dictionary are within the specified range."""
        lower, upper = self._get_lo_freq_range(is_resonator)
        return bool(lower <= frequency <= upper)

    def _validate_pulse_channel_if(self, frequency: float, is_resonator: bool) -> bool:
        """Validates that the pulse channel frequencies are within the specified range."""
        lower, upper = self._get_if_freq_range(is_resonator)
        return bool(lower <= np.abs(frequency) <= upper)

    def _get_lo_freq_range(self, is_resonator: bool) -> tuple[float, float]:
        """Returns the frequency range for the given type of channel."""
        return (
            self.qubit_lo_freq_limits if not is_resonator else self.resonator_lo_freq_limits
        )

    def _get_if_freq_range(self, is_resonator: bool) -> tuple[float, float]:
        """Returns the IF frequency range for the given type of channel."""
        return (
            self.qubit_if_freq_limits if not is_resonator else self.resonator_if_freq_limits
        )

    def _find_physical_channel_violations(self, physical_channels) -> list[str] | None:
        """Finds violations in the physical channels based on their baseband frequencies."""
        violations = []
        for physical_channel in physical_channels:
            data = self._physical_channel_data.get(physical_channel, None)
            if not data["is_valid"]:
                freq_range = self._get_lo_freq_range(data["is_resonator"])
                baseband_freq = data["baseband_frequency"]
                name = "Resonator" if data["is_resonator"] else "Qubit"
                name += f" {data['index']}"
                violations.append(
                    f"Physical channel for {name} with baseband frequency "
                    f"{baseband_freq} is out of the valid "
                    f"range {freq_range}."
                )
        return violations

    def _find_pulse_channel_violations(self, pulse_channels) -> list[str] | None:
        """Finds violations in the pulse channels based on their IF frequencies."""
        violations = []
        for pulse_channel in pulse_channels:
            data = self._pulse_channel_data[pulse_channel]

            # TODO: Add support for custom pulse channels COMPILER-698
            valid = data["is_valid"]

            if not valid:
                freq_range = self._get_if_freq_range(data["is_resonator"])
                if_freq = data["if_frequency"]
                index = self._physical_channel_data[data["physical_channel"]]["index"]
                name = "Resonator" if data["is_resonator"] else "Qubit"
                name += f" {index} {data['pulse_type']}"
                violations.append(
                    f"The IF of {name} pulse channel has a value {if_freq}, "
                    f"which is outside of the the valid range {freq_range}."
                )

        return violations

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager containing the results of the analysis.
        """

        active_channel_res = res_mgr.lookup_by_type(ActivePulseChannelResults)
        pulse_channels = active_channel_res.targets
        physical_channels = set(
            [
                self._pulse_channel_data[pulse_channel]["physical_channel"]
                for pulse_channel in pulse_channels
            ]
        )

        violations = self._find_physical_channel_violations(physical_channels)
        violations.extend(self._find_pulse_channel_violations(pulse_channels))

        if len(violations) > 0:
            raise ValueError(
                "Frequency validation of the hardware model against the target data failed "
                "with the following violations: \n" + "\n".join(violations)
            )
        return ir


class ReturnSanitisationValidation(ValidationPass):
    """Validates that the IR has a :class:`Return` instruction."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        returns = [inst for inst in ir.instructions if isinstance(inst, Return)]

        if not returns:
            raise ValueError("Could not find any return instructions.")
        elif len(returns) > 1:
            raise ValueError("Found multiple return instructions.")
        return ir


class RepeatSanitisationValidation(ValidationPass):
    """Checks if the builder has a :class:`Repeat` instruction and warns if none exists."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        for inst in ir.instructions:
            if isinstance(inst, Repeat):
                return ir

        log.warning("Could not find any repeat instructions.")
        return ir


PydReadoutValidation = ReadoutValidation
PydDynamicFrequencyValidation = DynamicFrequencyValidation
PydHardwareConfigValidity = HardwareConfigValidity
PydFrequencySetupValidation = FrequencySetupValidation
PydNoMidCircuitMeasurementValidation = NoMidCircuitMeasurementValidation
PydReturnSanitisationValidation = ReturnSanitisationValidation
PydRepeatSanitisationValidation = RepeatSanitisationValidation



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import abc
from typing import Optional

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.executables import Executable
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class BaseBackend(abc.ABC):
    """
    Converts an intermediate representation (IR) to code for a given target
    by selecting target-machine operations to implement for each instruction
    in the IR.
    """

    def __init__(self, model: None | QuantumHardwareModel):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = model

    @abc.abstractmethod
    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: Optional[ResultManager] = None,
        met_mgr: Optional[MetricsManager] = None,
        compiler_config: Optional[CompilerConfig] = None,
    ) -> Executable: ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.runtime.base import BaseRuntime as BaseRuntime
from qat.runtime.base import ResultsAggregator as ResultsAggregator
from qat.runtime.legacy import LegacyRuntime as LegacyRuntime
from qat.runtime.simple import SimpleRuntime as SimpleRuntime

DefaultRuntime = SimpleRuntime



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Optional, Union

import numpy as np
from compiler_config.config import InlineResultsProcessing

from qat.ir.instructions import (
    Assign,
    BinaryOperator,
    Delay,
    FrequencyShift,
    Instruction,
    InstructionBlock,
    Jump,
    Label,
    PhaseReset,
    PhaseShift,
    Repeat,
    Reset,
    ResultsProcessing,
    Return,
    Synchronize,
)
from qat.ir.measure import (
    Acquire,
    AcquireMode,
    MeasureBlock,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    acq_mode_process_axis,
)
from qat.ir.waveforms import Pulse, SampledWaveform
from qat.model.device import Component, DrivePulseChannel, PulseChannel, Qubit
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.utils.pydantic import QubitId, ValidatedList

log = get_default_logger()


class InstructionBuilder(ABC):
    def __init__(
        self,
        hardware_model: PhysicalHardwareModel,
        instructions: list[Instruction] = [],
    ):
        self.hw = hardware_model
        self._qubit_index_by_uuid = {
            qubit.uuid: idx for (idx, qubit) in hardware_model.qubits.items()
        }
        self._ir = InstructionBlock(instructions=instructions)
        self.compiled_shots: int | None = None
        self.shots: int | None = None

    @abstractmethod
    def X(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ): ...

    @abstractmethod
    def Y(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ): ...

    @abstractmethod
    def Z(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ): ...

    @abstractmethod
    def U(
        self,
        target: Qubit,
        theta: float,
        phi: float,
        lamb: float,
        pulse_channel: PulseChannel = None,
    ): ...

    @abstractmethod
    def swap(self, target: Qubit, destination: Qubit): ...

    def had(self, target: Qubit):
        return self.Z(target).Y(target, theta=np.pi / 2.0)

    def SX(self, target: Qubit):
        return self.X(target, theta=np.pi / 2.0)

    def SXdg(self, target: Qubit):
        return self.X(target, theta=-(np.pi / 2.0))

    def S(self, target: Qubit):
        return self.Z(target, theta=np.pi / 2.0)

    def Sdg(self, target: Qubit):
        return self.Z(target, theta=-(np.pi / 2))

    def T(self, target: Qubit):
        return self.Z(target, theta=np.pi / 4.0)

    def Tdg(self, target: Qubit):
        return self.Z(target, -(np.pi / 4.0))

    @abstractmethod
    def controlled(
        self, controllers: Union[Qubit, list[Qubit]], builder: InstructionBuilder
    ): ...

    def cX(self, controllers: Union[Qubit, list[Qubit]], target: Qubit, theta=np.pi):
        return self.controlled(controllers, self.X(target, theta=theta))

    def cY(self, controllers: Union[Qubit, list[Qubit]], target: Qubit, theta=np.pi):
        return self.controlled(controllers, self.Y(target, theta=theta))

    def cZ(self, controllers: Union[Qubit, list[Qubit]], target: Qubit, theta=np.pi):
        return self.controlled(controllers, self.Z(target, theta=theta))

    def cnot(self, control: Union[Qubit, list[Qubit]], target: Qubit):
        return self.cX(control, target, theta=np.pi)

    @abstractmethod
    def ccnot(self, controllers: list[Qubit], target: Qubit): ...

    def cswap(
        self, controllers: Union[Qubit, list[Qubit]], target: Qubit, destination: Qubit
    ):
        return self.controlled(controllers, self.swap(target, destination))

    @abstractmethod
    def ECR(self, control: Qubit, target: Qubit): ...

    def repeat(self, repeat_count: int):
        return self.add(Repeat(repeat_count=repeat_count))

    def returns(self, variables: list[str] = None):
        """Add return statement."""
        variables = variables if variables is not None else []
        return self.add(Return(variables=variables))

    def reset(
        self, targets: Union[PulseChannel, Qubit] | Iterable[Union[PulseChannel, Qubit]]
    ):
        if isinstance(targets, list):
            targets = set(targets)
        elif isinstance(targets, PulseChannel | Qubit):
            targets = {targets}
        else:
            raise TypeError(
                f"Invalid type, expected '(PulseChannel | Qubit | List[PulseChannel | Qubit])' but got {type(targets)}."
            )

        pulse_channel_ids = []
        for target in targets:
            if isinstance(target, Qubit):
                self.add(Reset(qubit_target=self._qubit_index_by_uuid[target.uuid]))
                pulse_channel_ids.extend([pc.uuid for pc in target.all_pulse_channels])
            else:
                pulse_channel_ids.append(target.uuid)

        for pulse_ch_id in pulse_channel_ids:
            self.add(PhaseReset(target=pulse_ch_id))

        return self

    def assign(self, name: str, value):
        return self.add(Assign(name=name, value=value))

    def jump(self, label: Union[str, Label], condition: BinaryOperator | None = None):
        return self.add(Jump(label=label, condition=condition))

    def results_processing(self, variable: str, res_format: InlineResultsProcessing):
        return self.add(ResultsProcessing(variable=variable, results_processing=res_format))

    def add(self, *instructions: Instruction, flatten: bool = False):
        """
        Add one or more instruction(s) into this builder. All methods should use this
        instead of accessing the instructions tree directly as it deals with composite instructions.
        """
        self._ir.add(*instructions, flatten=flatten)
        return self

    def __add__(self, other: QuantumInstructionBuilder):
        comp_builder = self.__class__(
            hardware_model=self.hw, instructions=self._ir.instructions.model_copy()
        )
        if isinstance(other, InstructionBuilder):
            comp_builder.add(*other._ir.instructions)
        else:
            raise TypeError(
                "Only another `{self.__class__.__name__}` can be added to this builder."
            )
        return comp_builder

    @staticmethod
    def constrain(angle: float):
        """
        Constrain the rotation angle to avoid redundant rotations around the Bloch sphere.
        """
        return (angle + np.pi) % (2 * np.pi) - np.pi

    @classmethod
    def _check_identity_operation(cls, f):
        """
        Wrapper method to constrain the rotation angle and to determine whether to avoid redundant rotations around the Bloch sphere.
        """

        def wrapper(self, target, theta=np.pi, **kwargs):
            theta = self.constrain(theta)
            return self if np.isclose(theta, 0) else f(self, target, theta, **kwargs)

        return wrapper

    @property
    def instructions(self):
        return self._ir.instructions

    @instructions.setter
    def instructions(self, instructions: ValidatedList[Instruction]):
        self._ir.instructions = instructions

    @property
    def number_of_instructions(self):
        return self._ir.number_of_instructions

    def __iter__(self):
        return self._ir.__iter__()

    def __reversed__(self):
        return self._ir.__reversed__()

    def flatten(self):
        """
        Flatten the instruction builder by removing nested structures like InstructionBlocks.
        """
        return self._ir.flatten()


class QuantumInstructionBuilder(InstructionBuilder):
    def pretty_print(self):
        output_str = ""
        for instr in self:
            targets = getattr(instr, "targets", None)
            if targets:
                format_target = ""
                for target in targets:
                    format_target = f"{instr.__class__.__name__} -> {self.hw._ids_to_pulse_channels[target]} "
                    device = self.hw._pulse_channel_ids_to_device[target]
                    if isinstance(device, Qubit):
                        format_target += f" @Q{self.hw._qubits_to_qubit_ids[device]}"
                    else:
                        qubit = self.hw._resonators_to_qubits[device]
                        format_target += f" @R{self.hw._qubits_to_qubit_ids[qubit]}"
                output_str += format_target

            else:
                output_str += str(instr)

            output_str += "\n"

        return output_str.rstrip()

    @InstructionBuilder._check_identity_operation
    def X(self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None):
        """
        Adds a gate that drives the qubit with a rotation angle `theta` to the builder.

        :param target: The qubit to be rotated.
        :param theta: The applied rotation angle.
        :param pulse_channel: The pulse channel the pulses get sent to.
        """
        if np.isclose(theta, np.pi / 2.0):
            return self.add(*self._hw_X_pi_2(target, pulse_channel))
        elif np.isclose(theta, -np.pi / 2.0):
            return self.add(
                *self._hw_Z(target, theta=np.pi, pulse_channel=pulse_channel),
                *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
                *self._hw_Z(target, theta=np.pi, pulse_channel=pulse_channel),
            )
        return self.U(
            target,
            theta=theta,
            phi=-np.pi / 2.0,
            lamb=np.pi / 2.0,
            pulse_channel=pulse_channel,
        )

    @InstructionBuilder._check_identity_operation
    def Y(self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None):
        """
        Adds a gate that drives the qubit with a rotation angle `theta` to the builder.

        :param target: The qubit to be rotated.
        :param theta: The applied rotation angle.
        :param pulse_channel: The pulse channel the pulses get sent to.
        """
        if np.isclose(theta, np.pi / 2.0):
            return self.add(
                *self._hw_Z(target, theta=-np.pi / 2.0, pulse_channel=pulse_channel),
                *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
                *self._hw_Z(target, theta=np.pi / 2.0, pulse_channel=pulse_channel),
            )
        elif np.isclose(theta, -np.pi / 2.0):
            return self.add(
                *self._hw_Z(target, theta=np.pi / 2.0, pulse_channel=pulse_channel),
                *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
                *self._hw_Z(target, theta=-np.pi / 2.0, pulse_channel=pulse_channel),
            )
        return self.U(target, theta=theta, phi=0.0, lamb=0.0, pulse_channel=pulse_channel)

    @InstructionBuilder._check_identity_operation
    def Z(self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None):
        """
        Adds a virtual gate that rotates the reference frame of the qubit with a phase `theta`
        to the builder.

        :param target: The qubit to be rotated.
        :param theta: The applied rotation angle.
        """
        if pulse_channel:
            log.warning(
                "Pulse channel in Z-gate will be ignored in the `QuantumInstructionBuilder`."
            )

        return self.add(*self._hw_Z(target=target, theta=theta))

    def U(
        self,
        target: Qubit,
        theta: float,
        phi: float,
        lamb: float,
        pulse_channel: PulseChannel = None,
    ):
        """
        Adds an arbitrary rotation around the Bloch sphere with 3 Euler angles
        to the builder (see https://doi.org/10.48550/arXiv.1707.03429).

        :param target: The qubit to be rotated.
        :param theta: Rotation angle.
        :param phi: Rotation angle.
        :param lamb: Rotation angle.
        :param pulse_channel: The pulse channel the pulses get sent to.
        """
        theta = self.constrain(theta)
        return self.add(
            *self._hw_Z(target, theta=lamb + np.pi, pulse_channel=pulse_channel),
            *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
            *self._hw_Z(target, theta=np.pi - theta, pulse_channel=pulse_channel),
            *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
            *self._hw_Z(target, theta=phi, pulse_channel=pulse_channel),
        )

    def ZX(self, target1: Qubit, target2: Qubit, theta: float = np.pi / 4.0):
        """
        Adds a two-qubit interaction gate exp(-i \theta Z x X) to the builder.

        :param target1: The qubit to which Z gets applied to.
        :param target2: The qubit to which X gets applied to.
        :param theta: The applied rotation angle.
        """
        theta = self.constrain(theta)
        if np.isclose(theta, 0):
            return []

        target1_id = self._qubit_index_by_uuid[target1.uuid]
        target2_id = self._qubit_index_by_uuid[target2.uuid]

        cr_pulse_channel = target1.cross_resonance_pulse_channels[target2_id]
        cr_cancellation_pulse_channel = target2.cross_resonance_cancellation_pulse_channels[
            target1_id
        ]

        if np.isclose(theta, np.pi / 4.0):
            return self.add(*self._hw_ZX_pi_4(target1, target2))
        elif np.isclose(theta, -np.pi / 4.0):
            return self.add(
                PhaseShift(targets=cr_pulse_channel.uuid, phase=np.pi),
                PhaseShift(targets=cr_cancellation_pulse_channel.uuid, phase=np.pi),
                *self._hw_ZX_pi_4(target1, target2),
                PhaseShift(targets=cr_pulse_channel.uuid, phase=np.pi),
                PhaseShift(targets=cr_cancellation_pulse_channel.uuid, phase=np.pi),
            )
        else:
            raise NotImplementedError(
                "Generic ZX gate not implemented yet! Please use `theta` = pi/4 or -pi/4."
            )

    def _hw_X_pi_2(
        self, target: Qubit, pulse_channel: DrivePulseChannel = None, amp_scale: float = 1.0
    ):
        pulse_channel = pulse_channel or target.drive_pulse_channel

        try:
            pulse_waveform = pulse_channel.pulse.waveform_type(
                **pulse_channel.pulse.model_dump()
            )
            pulse_waveform.amp *= amp_scale
        except AttributeError as e:
            raise ValueError(
                f"Pulse channel {pulse_channel} does not have a valid pulse calibration."
            ) from e

        return [Pulse(targets=pulse_channel.uuid, waveform=pulse_waveform)]

    def _hw_Z(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ):
        if theta == 0:
            return []

        # Rotate drive pulse channel of the qubit.
        pulse_channel = pulse_channel or target.drive_pulse_channel
        instr_collection = [PhaseShift(targets=pulse_channel.uuid, phase=theta)]
        # Rotate all cross resonance (cancellation) pulse channels pertaining to the qubit.
        qubit_id = self._qubit_index_by_uuid[target.uuid]
        if isinstance(pulse_channel, DrivePulseChannel):
            for (
                coupled_qubit_id,
                crc_pulse_channel,
            ) in target.cross_resonance_cancellation_pulse_channels.items():
                coupled_qubit = self.hw.qubit_with_index(coupled_qubit_id)
                cr_pulse_channel = coupled_qubit.cross_resonance_pulse_channels[qubit_id]

                instr_collection.append(
                    PhaseShift(targets=cr_pulse_channel.uuid, phase=theta)
                )
                instr_collection.append(
                    PhaseShift(targets=crc_pulse_channel.uuid, phase=theta)
                )

        return instr_collection

    def _hw_ZX_pi_4(self, target1: Qubit, target2: Qubit):
        target1_id = self._qubit_index_by_uuid[target1.uuid]
        target2_id = self._qubit_index_by_uuid[target2.uuid]

        target1_pulse_channel = target1.cross_resonance_pulse_channels[target2_id]
        target2_pulse_channel = target2.cross_resonance_cancellation_pulse_channels[
            target1_id
        ]

        if target1_pulse_channel is None or target2_pulse_channel is None:
            raise ValueError(
                f"Tried to perform cross resonance on {str(target2)} "
                f"that is not linked to {str(target1)}."
            )

        pulse = target1_pulse_channel.zx_pi_4_pulse.model_dump()
        waveform_type = target1_pulse_channel.zx_pi_4_pulse.waveform_type
        if pulse is None:
            raise ValueError(
                f"No `zx_pi_4_pulse` available on {target1} with index {target1_id}."
            )

        sync = Synchronize(
            targets=[
                pulse_ch.uuid
                for pulse_ch in target1.all_pulse_channels
                + target1.resonator.all_pulse_channels
                + target2.all_pulse_channels
                + target2.resonator.all_pulse_channels
            ]
        )

        return [
            sync,
            Pulse(target=target1_pulse_channel.uuid, waveform=waveform_type(**pulse)),
            Pulse(target=target2_pulse_channel.uuid, waveform=waveform_type(**pulse)),
            sync,
        ]

    def _generate_measure_acquire(
        self, target: Qubit, mode: AcquireMode, output_variable: str = None
    ):
        # Measure-related info.
        measure_channel = target.measure_pulse_channel
        measure_instruction = Pulse(
            targets=measure_channel.uuid,
            waveform=measure_channel.pulse.waveform_type(
                **measure_channel.pulse.model_dump()
            ),
        )

        # Acquire-related info.
        acquire_channel = target.resonator.acquire_pulse_channel
        acquire_duration = (
            measure_channel.pulse.width
            if acquire_channel.acquire.sync
            else acquire_channel.acquire.width
        )

        if acquire_channel.acquire.use_weights is False:
            filter = Pulse(
                waveform=SampledWaveform(samples=acquire_channel.acquire.weights),
                duration=acquire_duration,
                target=None,
            )
        else:
            filter = None

        acquire_instruction = Acquire(
            targets=acquire_channel.uuid,
            duration=acquire_duration,
            mode=mode,
            delay=acquire_channel.acquire.delay,
            filter=filter,
            output_variable=output_variable,
        )

        return [
            measure_instruction,
            acquire_instruction,
        ]

    def measure(
        self,
        targets: Qubit | set[Qubit],
        mode: AcquireMode = AcquireMode.INTEGRATOR,
        output_variable: str = None,
        sync_qubits: bool = True,
    ) -> QuantumInstructionBuilder:
        """
        Measure one or more qubits.

        :param targets: The qubit(s) to be measured.
        :param mode: The type of acquisition at the level of the control hardware.
        :param sync_qubits: Flag determining whether to align the measurements of all
                            qubits in `targets` or not. Sync between qubits is on by default.
        """
        if isinstance(targets, Qubit):
            targets = [targets]

        target_ids = {self._qubit_index_by_uuid[target.uuid] for target in targets}
        measure_block = MeasureBlock(qubit_targets=target_ids)

        all_pulse_channels = [
            pc.uuid for qubit in targets for pc in qubit.all_pulse_channels
        ]

        if sync_qubits:
            # Sync pulse channels of all pulse channels.
            measure_block.add(Synchronize(targets=all_pulse_channels))

        for qubit in targets:
            if not sync_qubits:
                qubit_pulse_channels = [pc.uuid for pc in qubit.all_pulse_channels]
                measure_block.add(Synchronize(targets=qubit_pulse_channels))

            output_var = output_variable or self._generate_output_variable(qubit)

            # Measure and acquire instructions for the qubit.
            measure, acquire = self._generate_measure_acquire(
                qubit, mode=mode, output_variable=output_var
            )
            duration = max(measure.duration, acquire.delay + acquire.duration)
            measure_block.add(measure, acquire)
            measure_block.duration = max(measure_block.duration, duration)

            if not sync_qubits:
                qubit_pulse_channels = [pc.uuid for pc in qubit.all_pulse_channels]
                measure_block.add(Synchronize(targets=qubit_pulse_channels))

        # Sync pulse channels of all pulse channels after measurement.
        if sync_qubits:
            measure_block.add(Synchronize(targets=all_pulse_channels))

        return self.add(measure_block)

    def _apply_down_conversion_and_take_time_mean(
        self,
        acquire_mode: AcquireMode,
        target: Qubit | set[Qubit],
        output_variable: str,
    ):
        """
        Applies down conversion and takes the mean of the acquired data.
        If the acquire mode is INTEGRATOR, down conversion is not needed, so it returns the builder without any changes

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        if acquire_mode == AcquireMode.INTEGRATOR:
            return self

        return self.post_processing(
            target, output_variable, PostProcessType.DOWN_CONVERT, ProcessAxis.TIME
        ).post_processing(target, output_variable, PostProcessType.MEAN, ProcessAxis.TIME)

    def _take_sequence_mean(
        self,
        acquire_mode: AcquireMode,
        target: Qubit | set[Qubit],
        output_variable: str,
    ):
        """
        Applies the mean post-processing to the acquired data along the sequence axis. If the acquire mode is SCOPE,
        the mean post-processing is not needed, so it returns the builder without any changes.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        if acquire_mode == AcquireMode.SCOPE:
            return self

        return self.post_processing(
            target, output_variable, PostProcessType.MEAN, ProcessAxis.SEQUENCE
        )

    def measure_single_shot_z(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure a single qubit along the z-axis.
        Execution results in an array of floats centred around 0, where anything above 0 shows a bias towards a
        +Z measurement, and below 0 shows a bias towards a -Z measurement.
        The array is an ndarray, where the first dimension is for each of the qubits measured, the second dimension is
        for each shot run.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        self.measure(target, acquire_mode, output_variable)
        self._apply_down_conversion_and_take_time_mean(
            acquire_mode, target, output_variable
        )
        return self.post_processing(
            target, output_variable, PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL
        )

    def measure_single_shot_signal(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure the signal of a single qubit.
        Execution results in an array of complex numbers that show the war signal output form the hardware.
        The array is an ndarray, where the first dimension is for each of the qubits measured, the second dimension is
        for each shot run.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        self.measure(target, acquire_mode, output_variable)
        return self._apply_down_conversion_and_take_time_mean(
            acquire_mode, target, output_variable
        )

    def measure_mean_z(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure a single qubit along the z-axis.
        Execution results in an array of floats centred around 0, where anything above 0 shows a bias towards a
        +Z measurement, and below 0 shows a bias towards a -Z measurement.
        Each entry is for each qubit measured, and is the mean of the results from `measure_single_shot_z`.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        self.measure(target, acquire_mode, output_variable)
        self._apply_down_conversion_and_take_time_mean(
            acquire_mode, target, output_variable
        )
        self._take_sequence_mean(acquire_mode, target, output_variable)
        return self.post_processing(
            target, output_variable, PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL
        )

    def measure_mean_signal(self, target: Qubit, output_variable: str = None):
        """
        Measure the signal of a single qubit.
        Execution results in an array of complex numbers that show the war signal output form the hardware.
        Each entry is for each qubit measured, and is the mean of the results from `measure_single_signal_shot`.

        :param target: The qubit to be measured.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[ProcessAxis.SEQUENCE]
        return self.measure(target, acquire_mode, output_variable)._take_sequence_mean(
            acquire_mode, target, output_variable
        )

    def measure_scope_mode(self, target: Qubit, output_variable: str = None):
        """
        Measure the scope mode

        :param target: The qubit to be measured.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        acquire_mode = acq_mode_process_axis[ProcessAxis.TIME]
        return self.measure(target, acquire_mode, output_variable).post_processing(
            target, output_variable, PostProcessType.DOWN_CONVERT, [ProcessAxis.TIME]
        )

    def measure_single_shot_binned(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure a single qubit along the z-axis.
        Execution results in an array of ±1s, where 1 indicates a +Z measurement, and -1 indicates a -Z measurement.
        It takes the results from `measure_single_shot_z` and applies a binning operation to round the results to ±1.
        The array is an ndarray, where the first dimension is for each of the qubits measured, the second dimension is
        for each shot run.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        return (
            self.measure(target, acq_mode_process_axis[axis], output_variable)
            ._apply_down_conversion_and_take_time_mean(
                acquire_mode, target, output_variable
            )
            .post_processing(
                target, output_variable, PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL
            )
            .post_processing(target, output_variable, PostProcessType.DISCRIMINATE)
        )

    def _generate_output_variable(self, target: Component):
        return "out_" + target.uuid + f"_{np.random.randint(np.iinfo(np.int32).max)}"

    def _find_valid_measure_block(self, target_ids: QubitId | set[QubitId]):
        """
        Finds the previous :class:`MeasureBlock` where the given qubits are not measured yet.

        :param target_ids: The indices of the qubits to be measured.
        """
        for instruction in reversed(self.instructions):
            if isinstance(instruction, MeasureBlock) and not any(
                target_ids & instruction.qubit_targets
            ):
                instruction.qubit_targets.update(target_ids)
                return instruction
        return None

    def post_processing(
        self,
        target: Qubit,
        output_variable: str,
        process_type: PostProcessType,
        axes: Optional[Union[ProcessAxis, list[ProcessAxis]]] = None,
        args=None,
    ):
        axes = axes if axes is not None else []
        args = args if args is not None else []

        # Default the mean z-map args if none supplied.
        if not any(args):
            if process_type == PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
                args = target.mean_z_map_args

            elif process_type == PostProcessType.DISCRIMINATE:
                args = [target.discriminator]

            elif process_type == PostProcessType.DOWN_CONVERT:
                resonator = target.resonator
                phys_channel = resonator.physical_channel
                bb = phys_channel.baseband
                measure_pulse_ch = resonator.measure_pulse_channel

                if measure_pulse_ch.fixed_if:
                    args = [bb.if_frequency, phys_channel.sample_time]
                else:
                    args = [
                        measure_pulse_ch.frequency - bb.frequency,
                        phys_channel.sample_time,
                    ]

        axes = axes if isinstance(axes, list) else [axes]

        return self.add(
            PostProcessing(
                output_variable=output_variable,
                process_type=process_type,
                axes=axes,
                args=args,
            )
        )

    def cnot(self, control: Qubit, target: Qubit):
        return (
            self.ECR(control, target)
            .X(control)
            .Z(control, theta=-np.pi / 2)
            .X(target, theta=-np.pi / 2)
        )

    def ECR(self, control: Qubit, target: Qubit):
        if not isinstance(control, Qubit) or not isinstance(target, Qubit):
            raise ValueError("The quantum targets of the ECR node must be qubits!")

        target_id = self._qubit_index_by_uuid[target.uuid]

        if not control.cross_resonance_pulse_channels.get(target_id, None):
            raise ValueError(
                f"Qubits {self._qubit_index_by_uuid[control.uuid]} and {target_id} are not coupled."
            )

        return (
            self.ZX(control, target, theta=np.pi / 4.0)
            .X(control, theta=np.pi)
            .ZX(control, target, theta=-np.pi / 4.0)
        )

    def swap(self, target: Qubit, destination: Qubit):
        raise NotImplementedError("Not available on this hardware model.")

    def pulse(self, **kwargs):
        return self.add(Pulse(**kwargs))

    def acquire(
        self,
        target: Qubit,
        delay: float = 1e-06,
        **kwargs,
    ):
        pulse_channel = target.acquire_pulse_channel

        if delay is None:
            kwargs["delay"] = pulse_channel.acquire.delay

        return self.add(Acquire(target=pulse_channel.uuid, **kwargs))

    def delay(self, target: Qubit | PulseChannel, duration: float):
        pulse_channel = target.drive_pulse_channel if isinstance(target, Qubit) else target
        return self.add(Delay(targets=pulse_channel.uuid, duration=duration))

    def synchronize(self, targets: Qubit | PulseChannel | list[Qubit | PulseChannel]):
        targets = targets if isinstance(targets, list) else [targets]

        pulse_channel_ids = set()
        for target in targets:
            if isinstance(target, PulseChannel):
                pulse_channel_ids.add(target.uuid)
            elif isinstance(target, Qubit):
                # TODO: At some point, we might want to implement (pulse channel)
                # getters at the qubit level just for the sake of conveniency. #326
                pulse_channel_ids.add(target.acquire_pulse_channel.uuid)
                pulse_channel_ids.add(target.measure_pulse_channel.uuid)
                qubit_pulse_channel_ids = [
                    pulse_channel.uuid for pulse_channel in target.all_pulse_channels
                ]
                pulse_channel_ids.update(qubit_pulse_channel_ids)
            else:
                raise TypeError(
                    "Please provide :class:`Qubit`s or :class:`PulseChannel`s as targets."
                )
        if len(pulse_channel_ids) > 1:
            return self.add(Synchronize(targets=pulse_channel_ids))
        else:
            log.warning(
                f"Ignored synchronisation of a single pulse channel with id {next(iter(pulse_channel_ids))}."
            )
            return self

    def controlled(
        self, controllers: Union[Qubit, list[Qubit]], builder: InstructionBuilder
    ):
        raise NotImplementedError("Not available on this hardware model.")

    def ccnot(self, controllers: list[Qubit], target: Qubit):
        raise NotImplementedError("Not available on this hardware model.")

    @InstructionBuilder._check_identity_operation
    def phase_shift(self, target: Qubit | PulseChannel, theta: float):
        if isinstance(target, Qubit):
            return self.add(
                PhaseShift(targets=target.drive_pulse_channel.uuid, phase=theta)
            )
        elif isinstance(target, PulseChannel):
            return self.add(PhaseShift(targets=target.uuid, phase=theta))
        else:
            raise TypeError(
                "Please provide a target that is either a `Qubit` or a `PulseChannel`."
            )

    def frequency_shift(self, target: Qubit, frequency):
        if isinstance(target, Qubit):
            return self.add(
                FrequencyShift(targets=target.drive_pulse_channel.uuid, frequency=frequency)
            )
        elif isinstance(target, PulseChannel):
            return self.add(FrequencyShift(targets=target.uuid, frequency=frequency))
        else:
            raise TypeError(
                "Please provide a target that is either a `Qubit` or a `PulseChannel`."
            )


PydQuantumInstructionBuilder = QuantumInstructionBuilder



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import json
import os
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime
from typing import Dict, List

import numpy as np
import regex
from qblox_instruments import (
    Cluster,
    DummyBinnedAcquisitionData,
    DummyScopeAcquisitionData,
)
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.backend.passes.purr.analysis import TILegalisationPass
from qat.purr.backends.live_devices import ControlHardware, Instrument, LivePhysicalBaseband
from qat.purr.backends.qblox.acquisition import Acquisition
from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.backends.qblox.config import (
    ModuleConfig,
    QbloxConfig,
    QcmConfigHelper,
    QcmRfConfigHelper,
    QrmConfigHelper,
    QrmRfConfigHelper,
    SequencerConfig,
)
from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.ir import Sequence
from qat.purr.backends.qblox.visualisation import plot_packages, plot_playback
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalChannel,
    PulseChannel,
    Qubit,
    Resonator,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxResonator(Resonator):
    """
    A hack around QBlox acquisition to use the same pulse channel
    for MeasurePulse and Acquire instructions.
    """

    def get_measure_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)

    def get_acquire_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)


class QbloxPhysicalBaseband(LivePhysicalBaseband):
    def __init__(
        self,
        id_,
        frequency,
        if_frequency,
        instrument: Instrument,
        slot_idx: int,
        config: QbloxConfig,
    ):
        super().__init__(id_, frequency, if_frequency, instrument)
        self.slot_idx = slot_idx
        self.config = config  # TBD, is it worth flattening ?


class QbloxPhysicalChannel(PhysicalChannel):
    baseband: QbloxPhysicalBaseband

    @property
    def slot_idx(self):
        return self.baseband.slot_idx

    @property
    def config(self):
        return self.baseband.config

    def build_resonator(self, resonator_id, *args, **kwargs):
        """Helper method to build a resonator with default channels."""
        kwargs.pop("fixed_if", None)
        reson = QbloxResonator(resonator_id, self)
        reson.create_pulse_channel(ChannelType.macq, *args, fixed_if=False, **kwargs)
        reson.default_pulse_channel_type = ChannelType.macq
        return reson

    def build_qubit(
        self,
        index,
        resonator,
        drive_freq,
        second_state_freq=None,
        channel_scale=(1.0e-8 + 0.0j),
        measure_amp: float = 1.0,
        fixed_drive_if=False,
        qubit_id=None,
    ):
        """
        Helper method tp build a qubit with assumed default values on the channels. Modelled after the live hardware.
        """
        qubit = Qubit(index, resonator, self, drive_amp=measure_amp, id_=qubit_id)
        qubit.create_pulse_channel(
            ChannelType.drive,
            frequency=drive_freq,
            scale=channel_scale,
            fixed_if=False,
        )

        qubit.create_pulse_channel(
            ChannelType.second_state, frequency=second_state_freq, scale=channel_scale
        )

        qubit.measure_acquire["delay"] = 151e-9  # TOF

        return qubit


class QbloxControlHardware(ControlHardware):
    _driver: Cluster

    def __init__(
        self,
        dev_id: str,
        name: str,
        address: str = None,
        dummy_cfg: Dict = None,
    ):
        super().__init__(id_=dev_id)
        self.name = name
        self.address = address
        self.dummy_cfg = dummy_cfg

        self.managed_mode = False
        self.dump_sequence = False
        self.plot_packages = False
        self.plot_acquisitions = False

        self._resources: Dict[Module, Dict[PulseChannel, Sequencer]] = {}

    def allocate_resources(self, package: QbloxPackage):
        target = package.target
        module_id = target.physical_channel.slot_idx - 1  # slot_idx is in range [1..20]
        module: Module = self._driver.modules[module_id]
        allocations = self._resources.setdefault(module, {})
        sequencer = allocations.get(target, None)
        if not sequencer:
            total = set(target.physical_channel.config.sequencers.keys())
            allocated = set([seq.seq_idx for seq in allocations.values()])

            available = total - allocated
            if not available:
                raise ValueError(f"No more available sequencers on module {module}")
            sequencer: Sequencer = module.sequencers[next(iter(available))]
            allocations[target] = sequencer

        log.debug(
            f"Sequencer {sequencer.seq_idx} in Module {module.slot_idx} will be running {package.target}"
        )
        return module, sequencer

    def configure(self, package: QbloxPackage, module: Module, sequencer: Sequencer):
        lo_freq, nco_freq = TILegalisationPass.decompose_freq(
            package.target.frequency, package.target
        )

        qblox_config = package.target.physical_channel.config

        # Customise Module config
        module_config: ModuleConfig = qblox_config.module
        if module_config.lo.out0_en:
            module_config.lo.out0_freq = lo_freq
        if module_config.lo.out1_en:
            module_config.lo.out1_freq = lo_freq
        if module_config.lo.out0_in0_en:
            module_config.lo.out0_in0_freq = lo_freq

        module_config.scope_acq.sequencer_select = sequencer.seq_idx
        module_config.scope_acq.trigger_mode_path0 = "sequencer"
        module_config.scope_acq.avg_mode_en_path0 = True
        module_config.scope_acq.trigger_mode_path1 = "sequencer"
        module_config.scope_acq.avg_mode_en_path1 = True

        # Customise Sequencer config
        sequencer_config: SequencerConfig = qblox_config.sequencers[sequencer.seq_idx]
        sequencer_config.nco.freq = nco_freq
        sequencer_config.nco.prop_delay_comp_en = True
        sequencer_config.square_weight_acq.integration_length = (
            package.sequencer_config.square_weight_acq.integration_length
        )
        sequencer_config.thresholded_acq.rotation = (
            package.sequencer_config.thresholded_acq.rotation
        )
        sequencer_config.thresholded_acq.threshold = (
            package.sequencer_config.thresholded_acq.threshold
        )

        log.debug(f"Configuring module {module}, sequencer {sequencer}")
        if module.is_qcm_type:
            if module.is_rf_type:
                config_helper = QcmRfConfigHelper(module_config, sequencer_config)
            else:
                config_helper = QcmConfigHelper(module_config, sequencer_config)
        elif module.is_qrm_type:
            if module.is_rf_type:
                config_helper = QrmRfConfigHelper(module_config, sequencer_config)
            else:
                config_helper = QrmConfigHelper(module_config, sequencer_config)
        else:
            raise ValueError(f"Unknown module type {module.module_type}")

        config_helper.configure(module, sequencer)

        if self.managed_mode:
            log.info("Managed mode enabled")
            config_helper.calibrate_mixer(module, sequencer)

    def _reset_connections(self):
        # TODO - Qblox bug: Hard reset clutters sequencer connections with conflicting defaults
        # TODO - This is a temporary workaround until Qblox fixes the issue

        modules = self._resources.keys() or self._driver.get_connected_modules().values()

        for m in modules:
            log.debug(f"Resetting sequencer connections for module {m.slot_idx}")
            m.disconnect_outputs()
            if m.is_qrm_type:
                m.disconnect_inputs()

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(
                name=self.name,
                identifier=self.address,
                dummy_cfg=self.dummy_cfg if self.address is None else None,
            )
            self._driver.reset()
            self._reset_connections()

        log.info(self._driver.get_system_status())
        self.is_connected = True

    def disconnect(self):
        if self._driver is not None:
            try:
                self._driver.close()
                self.is_connected = False
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument ID: {self.id} at: {self.address}\n{str(e)}"
                )

    def set_data(self, packages: List[QbloxPackage]):
        if self.plot_packages:
            plot_packages(packages)

        if self.dump_sequence:
            for pkg in packages:
                filename = f"schedules/target_{pkg.target.id}_@_{datetime.now().strftime('%m-%d-%Y_%H%M%S')}.json"
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, "w") as f:
                    f.write(json.dumps(asdict(pkg.sequence)))

        try:
            self._resources.clear()
            self._reset_connections()
            for pkg in packages:
                module, sequencer = self.allocate_resources(pkg)
                self.configure(pkg, module, sequencer)
                sequence = asdict(pkg.sequence)
                log.debug(f"Uploading sequence to {module}, sequencer {sequencer}")
                sequencer.sequence(sequence)
        except BaseException as e:
            self._reset_connections()
            self._resources.clear()
            raise e

    def start_playback(self, repetitions: int, repetition_time: float):
        if not any(self._resources):
            raise ValueError("No resources allocated. Install packages first")

        results: Dict[PulseChannel, List[Acquisition]] = defaultdict(list)
        try:
            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.sync_en(True)

            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.arm_sequencer()

            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.start_sequencer()

            for module, allocations in self._resources.items():
                if module.is_qrm_type:
                    for target, sequencer in allocations.items():
                        status_obj = sequencer.get_sequencer_status()
                        log.debug(f"Sequencer status - {sequencer}: {status_obj}")
                        if acquisitions := sequencer.get_acquisitions():
                            # TODO - 60 min tops, make it dynamic by involving flow-aware timeline duration
                            # Only wait if you sequencer is expected to have acquisitions
                            # TODO - Precise expectation of acquisitions should come from higher up
                            sequencer.get_acquisition_status(timeout=60)

                        for name in acquisitions:
                            sequencer.store_scope_acquisition(name)

                        # (re)fetch the lot after having stored scope acquisition
                        acquisitions = sequencer.get_acquisitions()

                        integ_length = sequencer.integration_length_acq()
                        start, end = (
                            0,
                            min(
                                integ_length,
                                Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS,
                            ),
                        )
                        for name, acquisition in acquisitions.items():
                            acquisition = Acquisition.model_validate(acquisition)
                            acquisition.name = name

                            scope_data = acquisition.acquisition.scope
                            scope_data.path0.data = scope_data.path0.data[start:end]
                            scope_data.path1.data = scope_data.path1.data[start:end]

                            integ_data = acquisition.acquisition.bins.integration
                            integ_data.path0 /= integ_length
                            integ_data.path1 /= integ_length

                            results[target].append(acquisition)

                        sequencer.delete_acquisition_data(all=True)

            if self.plot_acquisitions:
                plot_playback(results)

            return results
        finally:
            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.sync_en(False)

            self._reset_connections()
            self._resources.clear()

    def __getstate__(self) -> Dict:
        results = super(QbloxControlHardware, self).__getstate__()
        results["_driver"] = None
        results["_resources"] = {}
        return results

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._driver = None
        self._resources = {}
        self.is_connected = False


class DummyQbloxControlHardware(QbloxControlHardware):
    shot_pattern = regex.compile("jlt( +)R([0-9]+),([0-9]+),@(.*)\n")

    def _setup_dummy_scope_acq_data(self, module, sequencer: Sequencer, sequence: Sequence):
        shot_match = next(self.shot_pattern.finditer(sequence.program), None)
        avg_count = int(shot_match.group(3)) if shot_match else 1

        dummy_data = np.random.random(
            size=(Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS, 2)
        )
        dummy_data = [(iq[0], iq[1]) for iq in dummy_data]
        dummy_scope_acquisition_data = DummyScopeAcquisitionData(
            data=dummy_data, out_of_range=(False, False), avg_cnt=(avg_count, avg_count)
        )
        module.set_dummy_scope_acquisition_data(
            sequencer=sequencer.seq_idx, data=dummy_scope_acquisition_data
        )

    def _setup_dummy_binned_acq_data(
        self, module, sequencer: Sequencer, sequence: Sequence
    ):
        shot_match = next(self.shot_pattern.finditer(sequence.program), None)
        avg_count = int(shot_match.group(3)) if shot_match else 1

        for name, acquisition in sequence.acquisitions.items():
            dummy_binned_acquisition_data = [
                DummyBinnedAcquisitionData(
                    data=(np.random.random(), np.random.random()),
                    thres=np.random.choice(2),
                    avg_cnt=avg_count,
                )
            ] * acquisition["num_bins"]
            module.set_dummy_binned_acquisition_data(
                sequencer=sequencer.seq_idx,
                acq_index_name=name,
                data=dummy_binned_acquisition_data,
            )

    def _delete_acquisitions(self, sequencer):
        sequencer.delete_dummy_scope_acquisition_data()
        sequencer.delete_dummy_binned_acquisition_data()

    def set_data(self, packages: List[QbloxPackage]):
        super().set_data(packages)

        # Stage Scope and Acquisition data
        for module, allocations in self._resources.items():
            if module.is_qrm_type:
                for target, sequencer in allocations.items():
                    package = next((pkg for pkg in packages if pkg.target == target))
                    self._setup_dummy_scope_acq_data(module, sequencer, package.sequence)
                    self._setup_dummy_binned_acq_data(module, sequencer, package.sequence)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import numpy as np

from qat.core.pass_base import TransformPass
from qat.core.result_base import ResultManager
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.instructions import Sweep


class DesugaringPass(TransformPass):
    """Transforms syntactic sugars and implicit effects. Not that it applies here, but a
    classical example of a sugar syntax is the ternary expression
    :code:`"cond ? val1 : val2"` which is nothing more than an if else in disguise.

    The goal of this pass is to desugar any syntactic and semantic sugars. One of these
    sugars is iteration constructs such as :class:`Sweep` and :class:`Repeat`.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        for inst in ir.instructions:
            if isinstance(inst, Sweep):
                count = len(next(iter(inst.variables.values())))
                iter_name = f"sweep_{hash(inst)}"
                inst.variables[iter_name] = np.linspace(1, count, count, dtype=int).tolist()
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from numbers import Number
from typing import Dict, List, Union

import numpy as np

from qat.ir.measure import PostProcessing
from qat.purr.compiler.instructions import AcquireMode, PostProcessType, ProcessAxis

# TODO: move to QAT config?
UPCONVERT_SIGN = 1.0


def get_axis_map(mode: AcquireMode, response: np.ndarray) -> Dict[ProcessAxis, int]:
    """
    Given the acquisition mode, determine what each axis corresponds to.

    The way the results are returned are defined by the acquisition mode: this could be
    averaged over shots, averaged over time, or neither. We must determine how to unpack
    the results.

    :param mode: The acquisition mode.
    :param response: The response returned by the target machine.
    :returns: A dictionary containing the axis index for each type of :class:`ProcessAxis`.
    """

    match mode:
        case AcquireMode.SCOPE:
            return {ProcessAxis.TIME: -1}
        case AcquireMode.INTEGRATOR:
            return {ProcessAxis.SEQUENCE: -1}
        case AcquireMode.RAW:
            if response.ndim > 1:
                return {ProcessAxis.SEQUENCE: -2, ProcessAxis.TIME: -1}
            else:
                return {ProcessAxis.TIME: -1}
    raise NotImplementedError(f"The acquisition mode {mode} is not supported.")


def apply_post_processing(
    response: np.ndarray, post_processing: PostProcessing, axes: Dict[ProcessAxis, int]
) -> (np.ndarray, dict[ProcessAxis, int]):
    """
    Applies software post processing to the results.

    Uses the information in the :class:`PostProcessing` instruction to determine what method
    to  apply.

    :param response: Readout results from an execution engine.
    :param post_processing: The post processing instruction.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :returns: The processed results as an array and the axis map.
    """

    match post_processing.process_type:
        case PostProcessType.DOWN_CONVERT:
            _validate_down_convert_axis(post_processing.axes)
            return down_convert(response, axes, *post_processing.args)
        case PostProcessType.MEAN:
            return mean(response, axes, post_processing.axes)
        case PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
            return linear_map_complex_to_real(response, axes, *post_processing.args)
        case PostProcessType.DISCRIMINATE:
            return discriminate(response, axes, post_processing.args[0])

    raise NotImplementedError(
        f"Post processing type {post_processing.process_type} not implemented."
    )


def down_convert(
    response: np.ndarray,
    axes: Dict[ProcessAxis, int],
    frequency: float,
    dt: float,
):
    """
    Down-conversion of the readout signal.

    If down-conversion of the readout signal is not done on the hardware, is can be done via
    software using this method. Can only be done over the :attr:`ProcessAxis.TIME` axis.

    :param np.ndarray response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :type axes: dict[ProcessAxis, Int]
    :param float frequency: Down-conversion frequency
    :param float dt: The sampling rate for the readout
    :returns: The processed results as an array and the axis map.
    """

    axis = axes[ProcessAxis.TIME]

    # calculate the kernal used for down-conversion
    samples = response.shape[axis]
    t = np.linspace(0.0, dt * (samples - 1), samples)
    kernal = np.exp(-UPCONVERT_SIGN * 2.0j * np.pi * frequency * t)

    # multiply the response by the kernal over all dimensions.
    npaxis = [np.newaxis] * response.ndim
    npaxis[axis] = slice(None, None, None)
    result = response * kernal[tuple(npaxis)]
    return result, axes


def mean(
    response: np.ndarray,
    axes: Dict[ProcessAxis, int],
    target_axes: Union[ProcessAxis, List[ProcessAxis]],
):
    """
    Calculates the mean over the given axes.

    :param np.ndarray response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :type axes: dict[ProcessAxis, Int]
    :param target_axes: Which axis or axes should the mean be done over?
    :type target_axes: Union[ProcessAxis, list[ProcessAxis]]
    :returns: The processed results as an array and the axis map.
    """

    target_axes = target_axes if isinstance(target_axes, list) else [target_axes]
    axis_indices = tuple(axes[axis] for axis in axes if axis in target_axes)
    final_axes = _remove_axes(response.ndim, axis_indices, axes)
    final_data = np.mean(response, axis=axis_indices)
    return final_data, final_axes


def linear_map_complex_to_real(
    response: List[np.ndarray],
    axes: Dict[ProcessAxis, int],
    multiplier: Number,
    constant: Number,
):
    """
    Maps complex values onto a real z-projection using a provided linear mapping.

    :param np.ndarray response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :type axes: dict[ProcessAxis, Int]
    :param numbers.Number multiplier: Coeffecient for the linear map.
    :param numbers.Number constant: Constant for the linear map.
    :returns: The processed results as an array and the axis map.
    """

    return np.real(multiplier * response + constant), axes


def discriminate(response: np.ndarray, axes: Dict[ProcessAxis, int], threshold: float):
    """
    Discriminates a real value to a classical bit by comparison to a supplied
    discrimination threshold.

    :param np.ndarray response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :type axes: dict[ProcessAxis, Int]
    :param float threshold: The supplied discrimination threshold.
    :returns: The processed results as an array and the axis map.
    """
    return 2 * (response > threshold) - 1, axes


def _remove_axes(original_dims, removed_axis_indices, axis_locations):
    """
    Extracted from `purr/backends/utilities.py`.

    Returns the new axis map after axes have been removed, e.g., from calculating a mean.
    """

    # map original axis index to new axis index
    axis_map = {i: i for i in range(original_dims)}
    for r in removed_axis_indices:
        if r < 0:
            r = original_dims + r
        axis_map[r] = None
        for i in range(r + 1, original_dims):
            if axis_map[i] is not None:
                axis_map[i] -= 1
    axis_negative = {k: v < 0 for k, v in axis_locations.items()}
    new_axis_locations = {
        k: (original_dims + v if v < 0 else v) for k, v in axis_locations.items()
    }
    new_axis_locations = {
        k: axis_map[v] for k, v in new_axis_locations.items() if axis_map[v] is not None
    }
    new_dims = original_dims - len(removed_axis_indices)
    new_axis_locations = {
        k: (v - new_dims) if axis_negative[k] else v for k, v in new_axis_locations.items()
    }
    return new_axis_locations


def _validate_down_convert_axis(target_axis):
    if isinstance(target_axis, List):
        if len(target_axis) != 1:
            raise ValueError("Down conversion can only be done over a single axis.")
        target_axis = target_axis[0]

    if target_axis != ProcessAxis.TIME:
        raise ValueError("Down conversion can only be done over the time axis.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.engines.waveform_v1 import EchoEngine
from qat.model.convert_purr import convert_purr_echo_hw_to_pydantic
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import ExecutePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime import SimpleRuntime
from qat.runtime.results_pipeline import (
    get_experimental_results_pipeline,
)

log = get_default_logger()


class EchoExecutePipeline(UpdateablePipeline):
    """A pipeline that executes
    :class:`WaveformV1Executable <qat.backend.waveform_v1.executeable.WaveformV1Executable>`
    using the :class:`EchoEngine`.

    An engine cannot be provided to the pipeline, as the EchoEngine is used directly.

    .. warning::

        This pipeline is experimental and still in progress. Please use with caution.

        It is intended for executing compiled programs, and is not capable of compilation.
        Please use an appropriate compilation pipeline to prepare programs for execution.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> ExecutePipeline:
        """Constructs a pipeline equipped with the :class:`PydWaveformV1Backend`
        and :class:`EchoEngine`."""

        if engine is not None:
            log.warning(
                "An engine was provided to the EchoPipeline, but it will be ignored. "
                "The EchoEngine is used directly."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        pyd_model = convert_purr_echo_hw_to_pydantic(model)
        results_pipeline = get_experimental_results_pipeline(model, pyd_model)
        return ExecutePipeline(
            model=model,
            target_data=target_data,
            runtime=SimpleRuntime(engine=EchoEngine(), results_pipeline=results_pipeline),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import json
import os
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime
from functools import reduce
from itertools import groupby
from typing import Dict, List

from qblox_instruments import Cluster
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.backend.passes.purr.analysis import TILegalisationPass
from qat.backend.qblox.acquisition import Acquisition
from qat.backend.qblox.codegen import QbloxPackage
from qat.backend.qblox.config.constants import Constants
from qat.backend.qblox.config.helpers import (
    QcmConfigHelper,
    QcmRfConfigHelper,
    QrmConfigHelper,
    QrmRfConfigHelper,
)
from qat.backend.qblox.config.specification import (
    ModuleConfig,
    SequencerConfig,
)
from qat.backend.qblox.execution import QbloxExecutable
from qat.backend.qblox.visualisation import plot_executable
from qat.engines import NativeEngine
from qat.engines.qblox.instrument_base import CompositeInstrument, LeafInstrument
from qat.purr.backends.qblox.visualisation import plot_playback
from qat.purr.compiler.devices import (
    PulseChannel,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxInstrument(LeafInstrument):
    _driver: Cluster

    def __init__(
        self,
        dev_id: str,
        name: str,
        address: str = None,
        dummy_cfg: Dict = None,
    ):
        super().__init__(id=dev_id, name=name, address=address)
        self.dummy_cfg = dummy_cfg

        self.managed_mode = False
        self.dump_sequence = False
        self.plot_packages = False
        self.plot_acquisitions = False

        self._resources: Dict[Module, Dict[PulseChannel, Sequencer]] = {}

    def allocate_resources(self, target: PulseChannel):
        module_id = target.physical_channel.slot_idx - 1  # slot_idx is in range [1..20]
        module: Module = self._driver.modules[module_id]
        allocations = self._resources.setdefault(module, {})
        sequencer = allocations.get(target, None)
        if not sequencer:
            total = set(target.physical_channel.config.sequencers.keys())
            allocated = set([seq.seq_idx for seq in allocations.values()])

            available = total - allocated
            if not available:
                raise ValueError(f"No more available sequencers on module {module}")
            sequencer: Sequencer = module.sequencers[next(iter(available))]
            allocations[target] = sequencer

        log.debug(
            f"Sequencer {sequencer.seq_idx} in Module {module.slot_idx} will be running {target}"
        )
        return module, sequencer

    def configure(
        self,
        target: PulseChannel,
        package: QbloxPackage,
        module: Module,
        sequencer: Sequencer,
    ):
        lo_freq, nco_freq = TILegalisationPass.decompose_freq(target.frequency, target)

        qblox_config = target.physical_channel.config

        # Customise Module config
        module_config: ModuleConfig = qblox_config.module
        if module_config.lo.out0_en:
            module_config.lo.out0_freq = lo_freq
        if module_config.lo.out1_en:
            module_config.lo.out1_freq = lo_freq
        if module_config.lo.out0_in0_en:
            module_config.lo.out0_in0_freq = lo_freq

        module_config.scope_acq.sequencer_select = sequencer.seq_idx
        module_config.scope_acq.trigger_mode_path0 = "sequencer"
        module_config.scope_acq.avg_mode_en_path0 = True
        module_config.scope_acq.trigger_mode_path1 = "sequencer"
        module_config.scope_acq.avg_mode_en_path1 = True

        # Customise Sequencer config
        sequencer_config: SequencerConfig = qblox_config.sequencers[sequencer.seq_idx]
        sequencer_config.nco.freq = nco_freq
        sequencer_config.nco.prop_delay_comp_en = True
        sequencer_config.square_weight_acq.integration_length = (
            package.sequencer_config.square_weight_acq.integration_length
        )
        sequencer_config.thresholded_acq.rotation = (
            package.sequencer_config.thresholded_acq.rotation
        )
        sequencer_config.thresholded_acq.threshold = (
            package.sequencer_config.thresholded_acq.threshold
        )

        log.debug(f"Configuring module {module}, sequencer {sequencer}")
        if module.is_qcm_type:
            if module.is_rf_type:
                config_helper = QcmRfConfigHelper(module_config, sequencer_config)
            else:
                config_helper = QcmConfigHelper(module_config, sequencer_config)
        elif module.is_qrm_type:
            if module.is_rf_type:
                config_helper = QrmRfConfigHelper(module_config, sequencer_config)
            else:
                config_helper = QrmConfigHelper(module_config, sequencer_config)
        else:
            raise ValueError(f"Unknown module type {module.module_type}")

        config_helper.configure(module, sequencer)

        if self.managed_mode:
            log.info("Managed mode enabled")
            config_helper.calibrate_mixer(module, sequencer)

    def _reset_connections(self):
        # TODO - Qblox bug: Hard reset clutters sequencer connections with conflicting defaults
        # TODO - This is a temporary workaround until Qblox fixes the issue

        modules = self._resources.keys() or self._driver.get_connected_modules().values()

        for m in modules:
            log.debug(f"Resetting sequencer connections for module {m.slot_idx}")
            m.disconnect_outputs()
            if m.is_qrm_type:
                m.disconnect_inputs()

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(
                name=self.name,
                identifier=self.address,
                dummy_cfg=self.dummy_cfg if self.address is None else None,
            )
            self._driver.reset()
            self._reset_connections()

        log.info(self._driver.get_system_status())

    def disconnect(self):
        if self._driver is not None:
            try:
                self._driver.close()
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument ID: {self.id} at: {self.address}\n{str(e)}"
                )

    def setup(self, executable: QbloxExecutable):
        if self.plot_packages:
            plot_executable(executable)

        packages: Dict[PulseChannel, QbloxPackage] = executable.packages

        if self.dump_sequence:
            for target, pkg in packages.items():
                filename = f"schedules/target_{target.id}_@_{datetime.now().strftime('%m-%d-%Y_%H%M%S')}.json"
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, "w") as f:
                    f.write(json.dumps(asdict(pkg.sequence)))

        try:
            self._resources.clear()
            for target, pkg in packages.items():
                module, sequencer = self.allocate_resources(target)
                self.configure(target, pkg, module, sequencer)
                sequence = asdict(pkg.sequence)
                log.debug(f"Uploading sequence to {module}, sequencer {sequencer}")
                sequencer.sequence(sequence)
        except BaseException as e:
            self._reset_connections()
            self._resources.clear()
            raise e

    def playback(self):
        if not any(self._resources):
            raise ValueError("No resources allocated. Install packages first")

        results: Dict[PulseChannel, List[Acquisition]] = defaultdict(list)
        try:
            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.sync_en(True)

            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.arm_sequencer()

            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.start_sequencer()

            for module, allocations in self._resources.items():
                if module.is_qrm_type:
                    for target, sequencer in allocations.items():
                        # TODO - 60 min tops, make it dynamic by involving flow-aware timeline duration
                        sequencer.get_acquisition_status(timeout=60)
                        acquisitions = sequencer.get_acquisitions()

                        for name in acquisitions:
                            sequencer.store_scope_acquisition(name)

                        # (re)fetch the lot after having stored scope acquisition
                        acquisitions = sequencer.get_acquisitions()

                        integ_length = sequencer.integration_length_acq()
                        start, end = (
                            0,
                            min(
                                integ_length,
                                Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS,
                            ),
                        )
                        for name, acquisition in acquisitions.items():
                            acquisition = Acquisition.model_validate(acquisition)
                            acquisition.name = name

                            scope_data = acquisition.acq_data.scope
                            scope_data.i.data = scope_data.i.data[start:end]
                            scope_data.q.data = scope_data.q.data[start:end]

                            integ_data = acquisition.acq_data.bins.integration
                            integ_data.i /= integ_length
                            integ_data.q /= integ_length

                            results[target].append(acquisition)

                        sequencer.delete_acquisition_data(all=True)

            if self.plot_acquisitions:
                plot_playback(results)

            return results
        finally:
            for module, allocations in self._resources.items():
                for target, sequencer in allocations.items():
                    sequencer.sync_en(False)

            self._reset_connections()
            self._resources.clear()


class CompositeQbloxInstrument(CompositeInstrument):
    """
    For daisy-chained Qblox chassis.
    """

    pass


class QbloxEngine(NativeEngine):
    def __init__(self, instrument: QbloxInstrument | CompositeQbloxInstrument):
        self.instrument: QbloxInstrument | CompositeQbloxInstrument = instrument

    @staticmethod
    def combine_playbacks(playbacks: Dict[PulseChannel, List[Acquisition]]):
        """
        Combines acquisition objects from multiple acquire instructions in multiple readout targets.
        Notice that :meth:`groupby` preserves (original) relative order, which makes it honour
        the (sequential) lexicographical order of the loop nest:

        playback[target]["acq_0"] contains (potentially) a list of acquisitions collected in the same
        order as the order in which the packages were sent to the FPGA.

        Although acquisition names are enough for unicity in practice, the playback's structure
        distinguishes different (multiple) acquisitions per readout target, thus making it more robust.
        """

        playback: Dict[PulseChannel, Dict[str, Acquisition]] = {}
        for target, acquisitions in playbacks.items():
            groups_by_name = groupby(acquisitions, lambda acquisition: acquisition.name)
            playback[target] = {
                name: reduce(
                    lambda acq1, acq2: Acquisition.accumulate(acq1, acq2),
                    acqs,
                    Acquisition(),
                )
                for name, acqs in groups_by_name
            }

        return playback

    def execute(
        self, executable: QbloxExecutable
    ) -> Dict[PulseChannel, Dict[str, Acquisition]]:
        packages = executable.packages

        self.instrument.setup(packages)
        playbacks: Dict[PulseChannel, List[Acquisition]] = self.instrument.playback()
        playback: Dict[PulseChannel, Dict[str, Acquisition]] = self.combine_playbacks(
            playbacks
        )
        return playback



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


import numpy as np

from qat.executables import AcquireData, ChannelData, ChannelExecutable
from qat.utils.pydantic import ComplexNDArray


class WaveformV1ChannelData(ChannelData):
    """
    Contains the channel data for a :class:`WaveformV1Executable`.

    Stores the waveforms and acqusitions needed for execution. No control flow is possible.

    :param buffer: The waveform to be sent at each sample.
    :type buffer: list[complex]
    :param baseband_frequency: The frequency to be set for the baseband.
    :type baseband_frequency: float | None
    :param acquires: Acquire information needed for readouts.
    :type acquires: list[AcquireData]
    """

    buffer: ComplexNDArray = np.ndarray([])
    baseband_frequency: float | None = None
    acquires: list[AcquireData] = []

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.acquires != other.acquires:
            return False
        if self.baseband_frequency != other.baseband_frequency:
            return False
        if not np.all(self.buffer == other.buffer):
            return False
        return True


class WaveformV1Executable(ChannelExecutable):
    """
    A :class:`WaveformV1Executable` is an executable used in early iterations of QAT.

    This type of executable is composed of explicit waveforms that are compiled for each
    pulse channel, and summed to provide the waveform "buffer" for each physical channel.
    This early iteration can only instruct waveforms to be executed at a particular time;
    control flow is not possible. However, simple post-processing of results after execution
    is provided.

    :param channel_data: Stores the data required by the control hardware for each pulse
        channel.
    :type channel_data: dict[str, WaveformV1ChannelData]
    :param int shots: The number of times the program is executed.
    :param float repetition_time: The amount of time to wait between shots for the QPU to
        reset.
    :param PositiveInt compiled_shots: When the required number of shots exceeds the
        allowed amount by the target machine, shots can be batched into groups. This states
        how many shots to do in each batch.
    :param post_processing: Contains the post-processing information for each acquisition.
    :type post_processing: dict[str, list[PostProcessing]]
    :param results_processing: Contains the information for how results should be formatted.
    :type results_processing: dict[str, InlineResultsProcessing]
    :param assigns: Assigns results to given variables.
    :type assigns: list[Assign]
    :param returns: Which acquisitions/variables should be returned.
    :type returns: list[str]
    """

    channel_data: dict[str, WaveformV1ChannelData]
    repetition_time: float = 100e-6



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np

from qat.engines.native import NativeEngine
from qat.executables import AcquireData, Executable
from qat.purr.compiler.instructions import AcquireMode
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class ZeroEngine(NativeEngine):
    """An engine primarily used for testing that simply returns zero readout.

    The engine is designed to be generally applicable to most executables. It looks into the
    executable to retrieve all acquisitions, and creates appropiate readout data that is
    just filled with zeros.
    """

    def execute(self, package: Executable) -> dict[str, np.ndarray]:
        """Execute an :class:`Executable`, returning zeros for all readouts.

        :param package: The compiled executable containing acquisitions.
        :returns: The zero readout results.
        """
        shots = package.compiled_shots if package.compiled_shots else package.shots
        results = {}
        for acquire in package.acquires:
            results[acquire.output_variable] = np.zeros(
                readout_shape(acquire, shots), dtype=np.complex128
            )
        return results


def readout_shape(acquire: AcquireData, shots: int) -> tuple[int, ...]:
    """Generates the shape of the readout given the acquire information and number of shots.

    :param acquire: The acquisition information.
    :param shots: The number of shots being executed.
    :raises NotImplementedError: Only `AcquireMode.RAW`, `AcquireMode.INTEGRATOR` and
        `AcquireMode.SCOPE` are supported.
    :return: The shape of the returned array.
    """
    match acquire.mode:
        case AcquireMode.RAW:
            return (
                shots,
                acquire.length,
            )
        case AcquireMode.INTEGRATOR:
            return (shots,)
        case AcquireMode.SCOPE:
            return (acquire.length,)
    raise NotImplementedError(f"Acquire mode {acquire.mode} not currently supported.")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np

from qat.backend.waveform_v1 import WaveformV1Executable
from qat.engines.native import NativeEngine
from qat.purr.compiler.instructions import AcquireMode
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class EchoEngine(NativeEngine):
    """The :class:`EchoEngine` is a minimal execution engine primarily used for testing the
    compilation pipeline.

    It is not connected to any target machine such as live hardware or a simulator, and just
    simply "echos" back the buffers. Currently only accepts :class:`WaveformV1Executables`,
    but this may be changed in the future.
    """

    def execute(self, package: WaveformV1Executable):
        """
        Execute a :class:`WaveformV1Executable`.

        :param WaveformV1Executable package: The compiled executable.
        :returns: Execution results.
        :rtype: dict[str, np.ndarray]
        """
        shots = package.compiled_shots if package.compiled_shots else package.shots
        results = {}
        for channel_data in package.channel_data.values():
            buffer = np.asarray(channel_data.buffer)
            for acquire in channel_data.acquires:
                results[acquire.output_variable] = process_readout(
                    buffer[acquire.position : acquire.position + acquire.length],
                    shots,
                    acquire.mode,
                )
        return results


def process_readout(readout: np.ndarray, shots: int, mode: AcquireMode):
    """
    Processes a single readout into the expected format for a given acquire mode.

    For :attr:`AcquireMode.RAW`, this means repeating the readout for a given number of
    shots. The :attr:`AcquireMode.INTEGRATOR` emulates the averaging on hardware by
    taking the mean; note that this does not currently include down scaling. Finally, the
    :attr:`AcquireMode.SCOPE` simply returns the readout back in its current form.
    """
    match mode:
        case AcquireMode.RAW:
            return np.tile(readout, shots).reshape((shots, -1))
        case AcquireMode.INTEGRATOR:
            return np.tile(np.mean(readout, axis=0), shots)
        case AcquireMode.SCOPE:
            return readout
    raise NotImplementedError(f"Acquire mode {mode} not currently supported.")
